import json
from unittest.mock import Mock

from verse_sdk.contexts.span import SpanContext
from verse_sdk.contexts.types import EventMetadata


class TestSpanContext:
    def setup_method(self):
        self.mock_span = Mock()
        self.mock_span.add_event = Mock()
        self.mock_span.get_span_context.return_value.is_valid = True
        self.mock_span.record_exception = Mock()
        self.mock_span.set_attribute = Mock()
        self.mock_span.set_status = Mock()

    def test_event_basic(self):
        context = SpanContext(self.mock_span)
        result = context.event("test_event")

        assert result == context
        self.mock_span.add_event.assert_called_once_with(
            "test_event", {"level": "info"}
        )

    def test_event_with_custom_level(self):
        context = SpanContext(self.mock_span)
        result = context.event("error_event", level="error")

        assert result == context
        self.mock_span.add_event.assert_called_once_with(
            "error_event", {"level": "error"}
        )

    def test_event_with_metadata(self):
        context = SpanContext(self.mock_span)
        metadata: EventMetadata = {"key1": "value1", "key2": "value2"}
        result = context.event("test_event", metadata=metadata)

        assert result == context
        self.mock_span.add_event.assert_called_once_with(
            "test_event", {"level": "info", "metadata": json.dumps(metadata)}
        )

    def test_event_with_attrs(self):
        context = SpanContext(self.mock_span)
        result = context.event("test_event", custom_attr="value", another_attr=123)

        assert result == context
        self.mock_span.add_event.assert_called_once_with(
            "test_event", {"level": "info", "custom_attr": "value", "another_attr": 123}
        )

    def test_event_with_metadata_and_attrs(self):
        context = SpanContext(self.mock_span)
        metadata: EventMetadata = {"meta_key": "meta_value"}
        result = context.event("test_event", metadata=metadata, custom_attr="value")

        assert result == context
        self.mock_span.add_event.assert_called_once_with(
            "test_event",
            {"level": "info", "metadata": json.dumps(metadata), "custom_attr": "value"},
        )

    def test_event_with_none_metadata(self):
        context = SpanContext(self.mock_span)
        result = context.event("test_event", metadata=None)

        assert result == context
        self.mock_span.add_event.assert_called_once_with(
            "test_event", {"level": "info"}
        )

    def test_event_with_invalid_span(self):
        context = SpanContext(self.mock_span)
        self.mock_span.get_span_context.return_value.is_valid = False
        result = context.event("test_event")

        assert result == context
        self.mock_span.add_event.assert_not_called()



from collections.abc import AsyncIterator
from typing import Protocol


class AgentProtocol(Protocol):
    """Standard interface for all test agents."""

    def ask(self, question: str) -> str:
        """
        Ask the agent a question and return the response.

        Args:
            question: The question to ask the agent

        Returns:
            The agent's response as a string
        """
        ...

    async def stream(self, question: str) -> AsyncIterator[str]:
        """
        Ask the agent a question and stream the response.

        Args:
            question: The question to ask the agent

        Yields:
            Chunks of the agent's response
        """
        ...



from tests.e2e.runner.runner import E2ERunner
from tests.fixtures.datasource import FRIENDS_QUESTION
from verse_sdk import observe


class E2ERunnerSync(E2ERunner):
    @observe("e2e.runner.sync", type="trace")
    def run(self):
        response = self._agent.ask(FRIENDS_QUESTION)

        self.setup()
        self.assert_response_len(response)
        self.assert_validity()
        self.teardown()



import asyncio

from tests.e2e.runner.runner import E2ERunner
from tests.fixtures.datasource import STORY_QUESTION
from verse_sdk import observe


class E2ERunnerAsync(E2ERunner):
    @observe("e2e.runner.async", type="trace")
    def run(self):
        async def _run():
            chunks = []
            full_response = ""

            async for chunk in self._agent.stream(STORY_QUESTION):
                chunks.append(chunk)
                full_response = "".join(chunks)

            return full_response

        response = asyncio.run(_run())

        self.setup()
        self.assert_response_len(response)
        self.assert_validity()
        self.teardown()



from opentelemetry.sdk.trace.export import BatchSpanProcessor, SpanExporter

from .attribute import AttributeProcessor


class AttributeBatchSpanProcessor(BatchSpanProcessor):
    """
    Small wrapper so that we can injected project ID attributes per exporter rather than globally.
    Realistically, only Verse cares about project ID and attribute propagation.
    """

    def __init__(self, exporter: SpanExporter, project_id: str):
        """Initialize the processor with the exporter and project ID."""
        self.attribute_processor = AttributeProcessor(project_id)
        super().__init__(exporter)

    def on_start(self, span, parent_context=None):
        """Run attribute processor when span starts."""
        self.attribute_processor.on_start(span, parent_context)
        super().on_start(span, parent_context)

    def on_end(self, span):
        """Run attribute processor before exporting spans."""
        self.attribute_processor.on_end(span)
        super().on_end(span)



import json
import logging
from functools import cached_property
from typing import Any

from ..utils import format_hex_id
from .spec import OtelGenAISpanSpecification
from .types import (
    ERROR,
    INFO,
    OPTIONAL,
    RECOMMENDED,
    REQUIRED,
    WARNING,
    AttributeMetadata,
    AttributeSeverity,
    AttributeSpec,
    AttributeValidationDetail,
    AttributeValidationDetails,
)
from .utils import is_valid_type


class OtelGenAISpecCheck:
    def __init__(self, raw_span: Any):
        """Initialize the specification checker with a raw span.

        Args:
            raw_span: A span object (dict or JSON string) to validate against the
                OpenTelemetry GenAI semantic conventions specification.
        """
        self._status: str | None = None
        self._validation_details: AttributeValidationDetails = {}
        self._spec = OtelGenAISpanSpecification()
        self._load_span(raw_span)
        self._validate()

    @cached_property
    def could_improve(self) -> bool:
        """Check if the span has recommended attributes that are missing.

        Returns:
            bool: True if there are any warnings indicating missing recommended attributes.
        """
        return self._has_validation_detail_of_severity(WARNING)

    @cached_property
    def has_deprecations(self) -> bool:
        """Check if the span has any deprecations.

        Returns:
            bool: True if there are any deprecated attributes in use.
        """
        return self._has_validation_detail_of_severity(INFO)

    @cached_property
    def has_errors(self) -> bool:
        """Check if the span has any errors.

        Returns:
            bool: True if the span has validation errors or failed to load.
        """
        return self._span is None or self._has_validation_detail_of_severity(ERROR)

    def extract_messages(self) -> list[dict[str, Any]]:
        """Extract message history from the span.

        Returns:
            A list of message dictionaries. Currently returns an empty list.
        """
        return []

    def extract_metadata(self) -> AttributeMetadata:
        """Extract common metadata fields from the span.

        Returns:
            AttributeMetadata: A structured object containing environment, model information,
                session/user IDs, timing data, and trace context extracted from the span.
        """
        return AttributeMetadata(
            environment=self.get_attribute_value("environment", "dev"),
            model_provider=self.get_attribute_value("gen_ai.provider.name"),
            model=self.get_attribute_value("gen_ai.request.model"),
            project_id=self.get_attribute_value("project.id"),
            session_id=self.extract_session_id(),
            start_time=self.get_span_root_value("start_time"),
            end_time=self.get_span_root_value("end_time"),
            span_id=format_hex_id(self.get_span_context_value("span_id")),
            span_parent_id=format_hex_id(self.get_span_root_value("parent_id")),
            trace_id=format_hex_id(self.get_span_context_value("trace_id")),
            user_id=self.get_attribute_value("user.id"),
        )

    def extract_session_id(self) -> str:
        """Extract session ID from the span, checking multiple possible attribute names.

        Returns:
            The session ID from either 'gen_ai.conversation.id' or 'session.id' attributes,
            or None if neither is present.
        """
        return self.get_attribute_value(
            "gen_ai.conversation.id"
        ) or self.get_attribute_value("session.id")

    def get_attribute_value(self, name: str, default_value: Any = None) -> Any:
        """Get a span attribute value by name.

        Args:
            name: The attribute name to retrieve.
            default_value: Value to return if attribute is not found. Defaults to None.

        Returns:
            The attribute value if found, otherwise the default value.
        """
        return (
            self._span.get("attributes", {}).get(name, default_value)
            if self._span
            else default_value
        )

    def get_span_context_value(self, name: str, default_value: Any = None) -> Any:
        """Get a span context value by name.

        Args:
            name: The context field name to retrieve (e.g., 'span_id', 'trace_id').
            default_value: Value to return if field is not found. Defaults to None.

        Returns:
            The context value if found, otherwise the default value.
        """
        return (
            self._span.get("context", {}).get(name, default_value)
            if self._span
            else default_value
        )

    def get_span_root_value(self, name: str, default_value: Any = None) -> Any:
        """Get a root-level span field value by name.

        Args:
            name: The root field name to retrieve (e.g., 'start_time', 'end_time', 'parent_id').
            default_value: Value to return if field is not found. Defaults to None.

        Returns:
            The root field value if found, otherwise the default value.
        """
        return self._span.get(name, default_value) if self._span else default_value

    def is_attribute_deprecated(self, name: str) -> bool:
        """Check if an attribute is deprecated.

        Args:
            name: The attribute name to check.

        Returns:
            bool: True if the attribute is deprecated.
        """
        return self._get_validation_detail_of_severity(name, INFO) is not None

    def is_attribute_invalid(self, name: str) -> bool:
        """Check if an attribute is invalid.

        Args:
            name: The attribute name to check.

        Returns:
            bool: True if the attribute has validation errors.
        """
        return self._get_validation_detail_of_severity(name, ERROR) is not None

    def is_attribute_missing(self, name: str) -> bool:
        """Check if an attribute is missing.

        Args:
            name: The attribute name to check.

        Returns:
            bool: True if the attribute is recommended but missing.
        """
        return self._get_validation_detail_of_severity(name, WARNING) is not None

    def is_attribute_valid(self, name: str) -> bool:
        """Check the validity of an attribute by name.

        Args:
            name: The attribute name to check.

        Returns:
            bool: True if the attribute has no validation errors.
        """
        return self._get_validation_detail_of_severity(name, ERROR) is None

    def _discern_severity(
        self, attribute_spec: AttributeSpec, key: str, value: Any
    ) -> AttributeSeverity | None:
        requirement = attribute_spec.get("requirement", OPTIONAL)

        # Handle missing values
        if value is None:
            if requirement == REQUIRED:
                logging.error(f"Attribute {key} is required but not set")
                return ERROR
            elif requirement == RECOMMENDED:
                return WARNING
            return None

        # Handle deprecated attributes
        if attribute_spec.get("deprecated", False):
            logging.warning(f"Attribute {key} is deprecated")
            return INFO

        # Validate type
        expected_type = attribute_spec.get("type")
        if expected_type and not is_valid_type(value, expected_type):
            logging.error(
                f"Attribute {key} has invalid type. Expected {expected_type}, got {type(value).__name__}"
            )
            return ERROR

        # Validate enum values
        valid_values = attribute_spec.get("valid_values", [])
        if valid_values:
            if isinstance(value, list):
                if not all(item in valid_values for item in value):
                    logging.error(
                        f"Attribute {key} has invalid values. Expected one of {valid_values}"
                    )
                    return ERROR
            elif value not in valid_values:
                logging.error(
                    f"Attribute {key} has invalid value '{value}'. Expected one of {valid_values}"
                )
                return ERROR

        return None

    def _get_validation_detail_of_severity(
        self, key: str, severity: AttributeSeverity
    ) -> AttributeValidationDetail | None:
        match = next(
            (
                v
                for v in self._validation_details.values()
                if v["name"] == key and v["severity"] == severity
            ),
            None,
        )

        return match

    def _has_validation_detail_of_severity(self, severity: AttributeSeverity) -> bool:
        return any(v["severity"] == severity for v in self._validation_details.values())

    def _is_operation_specific_attribute(
        self, attribute_name: str, operation: str
    ) -> bool:
        if attribute_name.startswith("gen_ai.evaluation."):
            return operation not in ["", None]

        if attribute_name.startswith("gen_ai.embeddings."):
            return operation != "embeddings"

        if attribute_name.startswith("gen_ai.tool."):
            return operation != "execute_tool"

        return False

    def _is_provider_specific_attribute(
        self, attribute_name: str, provider: str
    ) -> bool:
        provider_prefixes = {
            "aws.bedrock.": "aws.bedrock",
            "openai.": "openai",
        }

        for prefix, required_provider in provider_prefixes.items():
            if attribute_name.startswith(prefix):
                return provider != required_provider

        return False

    def _load_span(self, raw_span: Any) -> None:
        try:
            self._span = json.loads(raw_span) if isinstance(raw_span, str) else raw_span

            if (
                self.get_span_context_value("trace_id") is None
                or self.get_span_context_value("span_id") is None
            ):
                logging.error("Trace ID or span ID is not set")
                self._status = "Trace ID or span ID is not set"
                self._span = None

        except (json.JSONDecodeError, TypeError, ValueError) as e:
            logging.error(f"Failed to load span: {e}")
            self._status = "Invalid span JSON"
            self._span = None

    def _push_validation_detail(
        self, attribute_spec: AttributeSpec, key: str, value: Any
    ) -> None:
        requirement = attribute_spec.get("requirement", OPTIONAL)
        severity = self._discern_severity(attribute_spec, key, value)

        if severity:
            self._validation_details[key] = AttributeValidationDetail(
                name=attribute_spec["name"],
                requirement=requirement,
                severity=severity,
                type=attribute_spec.get("type"),
                deprecated=attribute_spec.get("deprecated"),
                deprecated_message=attribute_spec.get("deprecated_message"),
                valid_values=attribute_spec.get("valid_values"),
            )

    def _should_validate_attribute(self, attribute_name: str, attrs: dict) -> bool:
        provider = attrs.get("gen_ai.provider.name", "")
        operation = attrs.get("gen_ai.operation.name", "")

        return not self._is_provider_specific_attribute(
            attribute_name,
            provider,
        ) and not self._is_operation_specific_attribute(
            attribute_name,
            operation,
        )

    def _validate(self) -> None:
        if self._span is None:
            return

        try:
            attrs = self._span.get("attributes", {})
            for attribute_spec in self._spec:
                name = attribute_spec["name"]

                if not self._should_validate_attribute(name, attrs):
                    continue

                matching = [
                    key for key in attrs if key.startswith(f"{name}.") or key == name
                ]

                if not matching:
                    self._push_validation_detail(attribute_spec, name, None)
                else:
                    for key in matching:
                        self._push_validation_detail(attribute_spec, key, attrs[key])

        except (AttributeError, KeyError) as e:
            logging.error(f"Validation error: {e}")
            self._status = "An error occurred while validating the span"



from __future__ import annotations

from unittest.mock import Mock, patch

import pytest

from verse_sdk.exporters.langfuse.exporter import LangfuseExporter


class TestLangfuseExporter:
    def setup_method(self):
        self.valid_config = {
            "public_key": "test_public_key",
            "secret_key": "test_secret_key",
        }

    def test_init_with_valid_config(self):
        exporter = LangfuseExporter(self.valid_config)

        assert exporter.config["public_key"] == "test_public_key"
        assert exporter.config["secret_key"] == "test_secret_key"

    def test_init_with_none_config(self):
        with patch.dict(
            "os.environ",
            {
                "LANGFUSE_PUBLIC_KEY": "env_public_key",
                "LANGFUSE_SECRET_KEY": "env_secret_key",
            },
        ):
            exporter = LangfuseExporter(None)

            assert exporter.config["public_key"] == "env_public_key"
            assert exporter.config["secret_key"] == "env_secret_key"

    def test_init_with_empty_config(self):
        with patch.dict(
            "os.environ",
            {
                "LANGFUSE_PUBLIC_KEY": "env_public_key",
                "LANGFUSE_SECRET_KEY": "env_secret_key",
            },
        ):
            exporter = LangfuseExporter({})

            assert exporter.config["public_key"] == "env_public_key"
            assert exporter.config["secret_key"] == "env_secret_key"

    def test_init_merges_env_and_config(self):
        with patch.dict(
            "os.environ",
            {
                "LANGFUSE_PUBLIC_KEY": "env_public_key",
                "LANGFUSE_SECRET_KEY": "env_secret_key",
                "LANGFUSE_HOST": "https://env.langfuse.com",
                "LANGFUSE_REGION": "us",
            },
        ):
            config = {
                "public_key": "config_public_key",
                "secret_key": "config_secret_key",
            }
            exporter = LangfuseExporter(config)

            assert exporter.config["public_key"] == "config_public_key"
            assert exporter.config["secret_key"] == "config_secret_key"
            assert exporter.config["host"] == "https://env.langfuse.com"
            assert exporter.config["region"] == "us"

    def test_validation_missing_public_key(self):
        config = {"secret_key": "test_secret_key"}

        with patch.dict("os.environ", {}, clear=True), pytest.raises(
            ValueError, match="Langfuse `public_key` must be set"
        ):
            LangfuseExporter(config)

    def test_validation_missing_secret_key(self):
        config = {"public_key": "test_public_key"}

        with patch.dict("os.environ", {}, clear=True), pytest.raises(
            ValueError, match="Langfuse `secret_key` must be set"
        ):
            LangfuseExporter(config)

    def test_validation_missing_both_keys(self):
        config = {}

        with patch.dict("os.environ", {}, clear=True), pytest.raises(
            ValueError,
            match="Langfuse `public_key` must be set\nLangfuse `secret_key` must be set",
        ):
            LangfuseExporter(config)

    def test_validation_with_none_values(self):
        config = {"public_key": None, "secret_key": None}

        with pytest.raises(
            ValueError,
            match="Langfuse `public_key` must be set\nLangfuse `secret_key` must be set",
        ):
            LangfuseExporter(config)

    def test_endpoint_with_custom_host(self):
        config = {
            "public_key": "test_public_key",
            "secret_key": "test_secret_key",
            "host": "https://custom.langfuse.com",
        }
        exporter = LangfuseExporter(config)

        assert (
            exporter.endpoint == "https://custom.langfuse.com/api/public/otel/v1/traces"
        )

    def test_endpoint_with_us_region(self):
        config = {
            "public_key": "test_public_key",
            "secret_key": "test_secret_key",
            "region": "US",
        }
        exporter = LangfuseExporter(config)

        assert (
            exporter.endpoint
            == "https://us.cloud.langfuse.com/api/public/otel/v1/traces"
        )

    def test_endpoint_with_eu_region(self):
        config = {
            "public_key": "test_public_key",
            "secret_key": "test_secret_key",
            "region": "EU",
        }
        with patch.dict("os.environ", {}, clear=True):
            exporter = LangfuseExporter(config)

            assert (
                exporter.endpoint
                == "https://cloud.langfuse.com/api/public/otel/v1/traces"
            )

    def test_endpoint_default_eu(self):
        config = {"public_key": "test_public_key", "secret_key": "test_secret_key"}
        with patch.dict("os.environ", {}, clear=True):
            exporter = LangfuseExporter(config)

            assert (
                exporter.endpoint
                == "https://cloud.langfuse.com/api/public/otel/v1/traces"
            )

    def test_endpoint_host_overrides_region(self):
        config = {
            "public_key": "test_public_key",
            "secret_key": "test_secret_key",
            "host": "https://custom.langfuse.com",
            "region": "US",
        }
        exporter = LangfuseExporter(config)

        assert (
            exporter.endpoint == "https://custom.langfuse.com/api/public/otel/v1/traces"
        )

    @patch("verse_sdk.exporters.langfuse.exporter.HTTPTraceExporter")
    @patch("verse_sdk.exporters.langfuse.exporter.ScopeFilterExporter")
    @patch("verse_sdk.exporters.langfuse.exporter.BatchSpanProcessor")
    def test_create_span_processor(
        self, mock_batch_processor, mock_scope_filter, mock_http_exporter
    ):
        with patch.dict("os.environ", {}, clear=True):
            config = {
                "public_key": "test_public_key",
                "secret_key": "test_secret_key",
                "host": "https://test.langfuse.com",
            }
            exporter = LangfuseExporter(config)

            mock_resource = Mock()
            mock_http_exporter.return_value = Mock()
            mock_scope_filter.return_value = Mock()
            mock_batch_processor.return_value = Mock()

            result = exporter.create_span_processor(mock_resource)

            mock_http_exporter.assert_called_once_with(
                endpoint="https://test.langfuse.com/api/public/otel/v1/traces",
                headers={
                    "authorization": "Basic dGVzdF9wdWJsaWNfa2V5OnRlc3Rfc2VjcmV0X2tleQ=="
                },
                timeout=10.0,
            )
            mock_scope_filter.assert_called_once_with(
                mock_http_exporter.return_value, exporter.config
            )
            mock_batch_processor.assert_called_once_with(mock_scope_filter.return_value)
            assert result == mock_batch_processor.return_value

    def test_env_method(self):
        with patch.dict(
            "os.environ",
            {
                "LANGFUSE_HOST": "https://test.langfuse.com",
                "LANGFUSE_PUBLIC_KEY": "env_public_key",
                "LANGFUSE_REGION": "us",
                "LANGFUSE_SECRET_KEY": "env_secret_key",
            },
        ):
            env_config = LangfuseExporter.env()

            assert env_config["host"] == "https://test.langfuse.com"
            assert env_config["public_key"] == "env_public_key"
            assert env_config["region"] == "us"
            assert env_config["secret_key"] == "env_secret_key"

    def test_env_method_missing_vars(self):
        with patch.dict("os.environ", {}, clear=True):
            env_config = LangfuseExporter.env()

            assert env_config["host"] is None
            assert env_config["public_key"] is None
            assert env_config["region"] is None
            assert env_config["secret_key"] is None

    def test_config_with_scopes(self):
        config = {
            "public_key": "test_public_key",
            "secret_key": "test_secret_key",
            "scopes": ["generation", "span"],
        }
        exporter = LangfuseExporter(config)

        assert exporter.config["scopes"] == ["generation", "span"]

    def test_config_with_all_options(self):
        config = {
            "public_key": "test_public_key",
            "secret_key": "test_secret_key",
            "host": "https://custom.langfuse.com",
            "region": "us",
            "scopes": ["generation"],
        }
        exporter = LangfuseExporter(config)

        assert exporter.config["public_key"] == "test_public_key"
        assert exporter.config["secret_key"] == "test_secret_key"
        assert exporter.config["host"] == "https://custom.langfuse.com"
        assert exporter.config["region"] == "us"
        assert exporter.config["scopes"] == ["generation"]

    def test_validation_logging(self):
        config = {"public_key": "test_public_key"}

        with patch.dict("os.environ", {}, clear=True), patch(
            "verse_sdk.exporters.langfuse.exporter.logging"
        ) as mock_logging, pytest.raises(
            ValueError, match="Langfuse `secret_key` must be set"
        ):
            LangfuseExporter(config)

        mock_logging.error.assert_called_once_with(
            "Langfuse validation errors: Langfuse `secret_key` must be set"
        )

    def test_validation_multiple_errors_logging(self):
        config = {}

        with patch.dict("os.environ", {}, clear=True), patch(
            "verse_sdk.exporters.langfuse.exporter.logging"
        ) as mock_logging, pytest.raises(
            ValueError, match="Langfuse `public_key` must be set"
        ):
            LangfuseExporter(config)

        mock_logging.error.assert_called_once_with(
            "Langfuse validation errors: Langfuse `public_key` must be set, Langfuse `secret_key` must be set"
        )



from __future__ import annotations

from unittest.mock import Mock, patch

import pytest
from opentelemetry.trace import SpanKind

from verse_sdk.contexts.utils import create_span_in_existing_trace


class TestCreateSpanInExistingTrace:
    def test_create_span_with_valid_hex_strings(self):
        trace_id_hex = "1234567890abcdef1234567890abcdef"
        parent_span_id_hex = "1234567890abcdef"

        with patch("verse_sdk.contexts.utils.trace") as mock_trace:
            mock_tracer = Mock()
            mock_span = Mock()
            mock_trace.get_tracer.return_value = mock_tracer
            mock_tracer.start_span.return_value = mock_span

            result = create_span_in_existing_trace(trace_id_hex, parent_span_id_hex)

            assert result == mock_span
            mock_trace.get_tracer.assert_called_once()
            mock_tracer.start_span.assert_called_once_with(
                "llm_generation",
                context=mock_trace.set_span_in_context.return_value,
                kind=SpanKind.INTERNAL,
            )

    def test_create_span_with_hex_prefixes(self):
        trace_id_hex = "0x1234567890abcdef1234567890abcdef"
        parent_span_id_hex = "0X1234567890abcdef"

        with patch("verse_sdk.contexts.utils.trace") as mock_trace:
            mock_tracer = Mock()
            mock_span = Mock()
            mock_trace.get_tracer.return_value = mock_tracer
            mock_tracer.start_span.return_value = mock_span

            result = create_span_in_existing_trace(trace_id_hex, parent_span_id_hex)

            assert result == mock_span
            mock_tracer.start_span.assert_called_once()

    def test_create_span_with_non_string_inputs(self):
        trace_id_hex = 0x1234567890ABCDEF1234567890ABCDEF
        parent_span_id_hex = 0x1234567890ABCDEF

        with patch("verse_sdk.contexts.utils.trace") as mock_trace:
            mock_tracer = Mock()
            mock_span = Mock()
            mock_trace.get_tracer.return_value = mock_tracer
            mock_tracer.start_span.return_value = mock_span

            result = create_span_in_existing_trace(trace_id_hex, parent_span_id_hex)

            assert result == mock_span
            mock_tracer.start_span.assert_called_once()

    def test_create_span_with_zero_values(self):
        trace_id_hex = "00000000000000000000000000000000"
        parent_span_id_hex = "0000000000000000"

        with patch("verse_sdk.contexts.utils.trace") as mock_trace:
            mock_tracer = Mock()
            mock_span = Mock()
            mock_trace.get_tracer.return_value = mock_tracer
            mock_tracer.start_span.return_value = mock_span

            result = create_span_in_existing_trace(trace_id_hex, parent_span_id_hex)

            assert result == mock_span
            mock_tracer.start_span.assert_called_once()

    def test_create_span_with_max_values(self):
        trace_id_hex = "ffffffffffffffffffffffffffffffff"
        parent_span_id_hex = "ffffffffffffffff"

        with patch("verse_sdk.contexts.utils.trace") as mock_trace:
            mock_tracer = Mock()
            mock_span = Mock()
            mock_trace.get_tracer.return_value = mock_tracer
            mock_tracer.start_span.return_value = mock_span

            result = create_span_in_existing_trace(trace_id_hex, parent_span_id_hex)

            assert result == mock_span
            mock_tracer.start_span.assert_called_once()

    def test_create_span_span_context_creation(self):
        """Test that SpanContext is created with correct parameters"""
        trace_id_hex = "1234567890abcdef1234567890abcdef"
        parent_span_id_hex = "1234567890abcdef"

        with patch("verse_sdk.contexts.utils.trace") as mock_trace, patch(
            "verse_sdk.contexts.utils.SpanContext"
        ) as mock_span_context_class, patch(
            "verse_sdk.contexts.utils.TraceFlags"
        ) as mock_trace_flags_class:
            mock_tracer = Mock()
            mock_span = Mock()
            mock_trace.get_tracer.return_value = mock_tracer
            mock_tracer.start_span.return_value = mock_span

            mock_span_context = Mock()
            mock_span_context_class.return_value = mock_span_context

            create_span_in_existing_trace(trace_id_hex, parent_span_id_hex)

            # Verify SpanContext was created with correct parameters
            expected_trace_id = int(trace_id_hex, 16)
            expected_span_id = int(parent_span_id_hex, 16)

            mock_span_context_class.assert_called_once_with(
                trace_id=expected_trace_id,
                span_id=expected_span_id,
                is_remote=True,
                trace_flags=mock_trace_flags_class.return_value,
            )

    def test_create_span_trace_flags_setting(self):
        """Test that TraceFlags is set correctly"""
        trace_id_hex = "1234567890abcdef1234567890abcdef"
        parent_span_id_hex = "1234567890abcdef"

        with patch("verse_sdk.contexts.utils.trace") as mock_trace, patch(
            "verse_sdk.contexts.utils.TraceFlags"
        ) as mock_trace_flags_class:
            mock_tracer = Mock()
            mock_span = Mock()
            mock_trace.get_tracer.return_value = mock_tracer
            mock_tracer.start_span.return_value = mock_span

            create_span_in_existing_trace(trace_id_hex, parent_span_id_hex)

            # Verify TraceFlags was created with correct value
            mock_trace_flags_class.assert_called_once_with(0x01)

    def test_create_span_non_recording_span_creation(self):
        """Test that NonRecordingSpan is created correctly"""
        trace_id_hex = "1234567890abcdef1234567890abcdef"
        parent_span_id_hex = "1234567890abcdef"

        with patch("verse_sdk.contexts.utils.trace") as mock_trace, patch(
            "verse_sdk.contexts.utils.SpanContext"
        ) as mock_span_context_class:
            mock_tracer = Mock()
            mock_span = Mock()
            mock_trace.get_tracer.return_value = mock_tracer
            mock_tracer.start_span.return_value = mock_span

            mock_span_context = Mock()
            mock_span_context_class.return_value = mock_span_context

            create_span_in_existing_trace(trace_id_hex, parent_span_id_hex)

            # Verify NonRecordingSpan was created
            mock_trace.NonRecordingSpan.assert_called_once_with(mock_span_context)

    def test_create_span_context_setting(self):
        """Test that span context is set correctly"""
        trace_id_hex = "1234567890abcdef1234567890abcdef"
        parent_span_id_hex = "1234567890abcdef"

        with patch("verse_sdk.contexts.utils.trace") as mock_trace:
            mock_tracer = Mock()
            mock_span = Mock()
            mock_trace.get_tracer.return_value = mock_tracer
            mock_tracer.start_span.return_value = mock_span

            mock_parent_span = Mock()
            mock_trace.NonRecordingSpan.return_value = mock_parent_span

            create_span_in_existing_trace(trace_id_hex, parent_span_id_hex)

            # Verify set_span_in_context was called
            mock_trace.set_span_in_context.assert_called_once_with(mock_parent_span)

    def test_create_span_with_invalid_hex_raises_error(self):
        trace_id_hex = "invalid_hex_string"
        parent_span_id_hex = "1234567890abcdef"

        with pytest.raises(
            ValueError, match="invalid literal for int\\(\\) with base 16"
        ):
            create_span_in_existing_trace(trace_id_hex, parent_span_id_hex)

    def test_create_span_with_empty_strings_raises_error(self):
        trace_id_hex = ""
        parent_span_id_hex = ""

        with pytest.raises(
            ValueError, match="invalid literal for int\\(\\) with base 16"
        ):
            create_span_in_existing_trace(trace_id_hex, parent_span_id_hex)

    def test_create_span_with_none_inputs_raises_error(self):
        with pytest.raises(
            ValueError, match="invalid literal for int\\(\\) with base 16: 'None'"
        ):
            create_span_in_existing_trace(None, None)

    def test_create_span_span_kind_is_internal(self):
        trace_id_hex = "1234567890abcdef1234567890abcdef"
        parent_span_id_hex = "1234567890abcdef"

        with patch("verse_sdk.contexts.utils.trace") as mock_trace:
            mock_tracer = Mock()
            mock_span = Mock()
            mock_trace.get_tracer.return_value = mock_tracer
            mock_tracer.start_span.return_value = mock_span

            create_span_in_existing_trace(trace_id_hex, parent_span_id_hex)

            call_args = mock_tracer.start_span.call_args
            assert call_args[1]["kind"] == SpanKind.INTERNAL

    def test_create_span_name_is_llm_generation(self):
        trace_id_hex = "1234567890abcdef1234567890abcdef"
        parent_span_id_hex = "1234567890abcdef"

        with patch("verse_sdk.contexts.utils.trace") as mock_trace:
            mock_tracer = Mock()
            mock_span = Mock()
            mock_trace.get_tracer.return_value = mock_tracer
            mock_tracer.start_span.return_value = mock_span

            create_span_in_existing_trace(trace_id_hex, parent_span_id_hex)

            call_args = mock_tracer.start_span.call_args
            assert call_args[0][0] == "llm_generation"

    def test_create_span_with_mixed_case_hex_prefixes(self):
        trace_id_hex = "0X1234567890abcdef1234567890abcdef"
        parent_span_id_hex = "0x1234567890abcdef"

        with patch("verse_sdk.contexts.utils.trace") as mock_trace:
            mock_tracer = Mock()
            mock_span = Mock()
            mock_trace.get_tracer.return_value = mock_tracer
            mock_tracer.start_span.return_value = mock_span

            result = create_span_in_existing_trace(trace_id_hex, parent_span_id_hex)

            assert result == mock_span
            mock_tracer.start_span.assert_called_once()

    def test_create_span_hex_conversion_edge_cases(self):
        test_cases = [
            ("0x0", "0x0"),
            ("0X0", "0X0"),
            ("0", "0"),
            ("1", "1"),
            ("a", "b"),
        ]

        for trace_id, parent_span_id in test_cases:
            with patch("verse_sdk.contexts.utils.trace") as mock_trace:
                mock_tracer = Mock()
                mock_span = Mock()
                mock_trace.get_tracer.return_value = mock_tracer
                mock_tracer.start_span.return_value = mock_span

                result = create_span_in_existing_trace(trace_id, parent_span_id)

                assert result == mock_span
                mock_tracer.start_span.assert_called_once()

    def test_create_span_return_type(self):
        trace_id_hex = "1234567890abcdef1234567890abcdef"
        parent_span_id_hex = "1234567890abcdef"

        with patch("verse_sdk.contexts.utils.trace") as mock_trace:
            mock_tracer = Mock()
            mock_span = Mock()
            mock_trace.get_tracer.return_value = mock_tracer
            mock_tracer.start_span.return_value = mock_span

            result = create_span_in_existing_trace(trace_id_hex, parent_span_id_hex)

            assert result == mock_span
            assert hasattr(result, "__class__")



import json
import os
from typing import Any

from google import genai
from google.genai import types as genai_types

from tests.fixtures.datasource import datasource


class GoogleAgent:
    """Agent for interacting with Google's Gemini API with function calling support."""

    DEFAULT_MODEL = "gemini-2.0-flash-exp"
    MAX_ITERATIONS = 3

    def __init__(self, api_key: str | None = None):
        """
        Initialize the Google Agent.

        Args:
            api_key: Google API key. If not provided, uses GOOGLE_API_KEY env variable.
        """
        self.client = genai.Client(api_key=api_key or os.getenv("GOOGLE_API_KEY"))
        self.model_name = self.DEFAULT_MODEL
        self.tools = self._create_tools()
        self.function_map = self._create_function_map()

    def _create_tools(self) -> list[genai_types.Tool]:
        """Convert datasource functions to Google GenAI tools."""
        function_declarations = [
            genai_types.FunctionDeclaration.from_callable(
                callable=func, client=self.client
            )
            for func in datasource
        ]
        return [genai_types.Tool(function_declarations=function_declarations)]

    def _create_function_map(self) -> dict[str, callable]:
        """Create a mapping of function names to callable functions."""
        return {func.__name__: func for func in datasource}

    def _execute_function_call(
        self, func_call: genai_types.FunctionCall
    ) -> genai_types.Part:
        """
        Execute a single function call and return the response part.

        Args:
            func_call: The function call to execute

        Returns:
            Part containing the function response
        """
        func_name = func_call.name
        func_args = func_call.args

        if isinstance(func_args, str):
            try:
                func_args = json.loads(func_args)
            except json.JSONDecodeError:
                func_args = {}

        if func_name not in self.function_map:
            return genai_types.Part(
                function_response=genai_types.FunctionResponse(
                    name=func_name,
                    response={"error": f"Function '{func_name}' not found"},
                )
            )

        try:
            result = (
                self.function_map[func_name](**func_args)
                if isinstance(func_args, dict)
                else self.function_map[func_name](func_args)
            )
            return genai_types.Part(
                function_response=genai_types.FunctionResponse(
                    name=func_name,
                    response=result,
                )
            )
        except Exception as e:
            return genai_types.Part(
                function_response=genai_types.FunctionResponse(
                    name=func_name,
                    response={"error": str(e)},
                )
            )

    def _extract_parts(self, content: Any) -> tuple[list, list]:
        """
        Extract function calls and text from response content.

        Returns:
            Tuple of (function_calls, text_parts)
        """
        parts = content.parts if hasattr(content, "parts") else []
        function_calls = []
        text_parts = []

        for part in parts:
            if hasattr(part, "function_call") and part.function_call:
                function_calls.append(part.function_call)
            elif hasattr(part, "text") and part.text:
                text_parts.append(part.text)

        return function_calls, text_parts

    def ask(self, question: str) -> list[Any]:
        """
        Ask a question and get a response, handling function calls automatically.

        Args:
            question: The question to ask

        Returns:
            List of response candidates
        """
        contents = [
            {
                "role": "user",
                "parts": [{"text": question}],
            }
        ]

        config = {
            "max_output_tokens": 500,
            "temperature": 0.7,
            "tools": self.tools,
            "tool_config": genai_types.ToolConfig(
                function_calling_config=genai_types.FunctionCallingConfig(mode="ANY")
            ),
        }

        for _ in range(self.MAX_ITERATIONS):
            response = self.client.models.generate_content(
                model=self.model_name,
                contents=contents,
                config=config,
            )

            if not response.candidates or not response.candidates[0].content:
                break

            content = response.candidates[0].content
            function_calls = self._extract_parts(content)

            if not function_calls:
                return response.candidates

            contents.append(
                {
                    "role": "model",
                    "parts": [
                        genai_types.Part(function_call=fc) for fc in function_calls
                    ],
                }
            )

            for func_call in function_calls:
                function_response = self._execute_function_call(func_call)
                contents.append(
                    {
                        "role": "user",
                        "parts": [function_response],
                    }
                )

        return response.candidates

    async def stream(self, question: str):
        """
        Stream responses from Google GenAI with tool call support.

        Args:
            question: The question to ask

        Yields:
            Text chunks from the streaming response
        """
        contents = [
            {
                "role": "user",
                "parts": [{"text": question}],
            }
        ]

        config = {
            "max_output_tokens": 500,
            "temperature": 0.7,
            "tools": self.tools,
            "tool_config": genai_types.ToolConfig(
                function_calling_config=genai_types.FunctionCallingConfig(mode="ANY")
            ),
        }

        for _ in range(self.MAX_ITERATIONS):
            stream = self.client.models.generate_content_stream(
                model=self.model_name,
                contents=contents,
                config=config,
            )

            full_response = None
            accumulated_text = []
            function_calls = []

            for chunk in stream:
                if (
                    hasattr(chunk, "candidates")
                    and chunk.candidates
                    and hasattr(chunk.candidates[0], "content")
                    and chunk.candidates[0].content
                ):
                    candidate = chunk.candidates[0]
                    content = candidate.content
                    if hasattr(content, "parts") and content.parts:
                        for part in content.parts:
                            if hasattr(part, "text") and part.text:
                                accumulated_text.append(part.text)
                                yield part.text
                            elif hasattr(part, "function_call") and part.function_call:
                                function_calls.append(part.function_call)
                    full_response = chunk

            if (
                function_calls
                and full_response
                and hasattr(full_response, "candidates")
                and full_response.candidates
                and hasattr(full_response.candidates[0], "content")
                and full_response.candidates[0].content
            ):
                candidate = full_response.candidates[0]
                content = candidate.content
                extracted_calls, _ = self._extract_parts(content)
                if extracted_calls:
                    contents.append(
                        {
                            "role": "model",
                            "parts": [
                                genai_types.Part(function_call=fc)
                                for fc in extracted_calls
                            ],
                        }
                    )

                    for func_call in extracted_calls:
                        function_response = self._execute_function_call(func_call)
                        contents.append(
                            {
                                "role": "user",
                                "parts": [function_response],
                            }
                        )
                    continue

            break



from unittest.mock import Mock

from verse_sdk.integrations.litellm.utils import (
    calculate_total_tokens,
    get_completition_choices,
    get_finish_reason,
    get_operation_type,
    get_stream_text,
    get_trace_metadata,
    get_usage,
    get_usage_from_response,
)


class TestCalculateTotalTokens:
    def test_both_tokens_provided(self):
        assert calculate_total_tokens(100, 50) == 150

    def test_only_prompt_tokens(self):
        assert calculate_total_tokens(100, None) == 100

    def test_only_completion_tokens(self):
        assert calculate_total_tokens(None, 50) == 50

    def test_both_none(self):
        assert calculate_total_tokens(None, None) is None

    def test_zero_tokens(self):
        assert (
            calculate_total_tokens(0, 0)
            == calculate_total_tokens(0, None)
            == calculate_total_tokens(None, 0)
            == 0
        )


class TestGetCompletionChoices:
    def test_object_with_choices_attribute(self):
        mock_data = Mock()
        mock_data.choices = [{"message": "test1"}, {"message": "test2"}]
        assert get_completition_choices(mock_data) == mock_data.choices

    def test_dict_with_choices_key(self):
        data = {"choices": [{"message": "test1"}, {"message": "test2"}]}
        assert get_completition_choices(data) == [
            {"message": "test1"},
            {"message": "test2"},
        ]

    def test_empty_choices_list(self):
        mock_data = Mock()
        mock_data.choices = []
        assert get_completition_choices(mock_data) == []

    def test_no_choices_attribute_or_key(self):
        mock_data = Mock()
        del mock_data.choices
        assert get_completition_choices(mock_data) == []

    def test_dict_without_choices_key(self):
        data = {"other_key": "value"}
        assert get_completition_choices(data) == []

    def test_none_data(self):
        assert get_completition_choices(None) == []

    def test_string_data(self):
        assert get_completition_choices("some string") == []


class TestGetFinishReason:
    def test_response_with_choices_and_finish_reason(self):
        mock_choice = Mock()
        mock_choice.finish_reason = "stop"

        mock_response = Mock()
        mock_response.choices = [mock_choice]
        assert get_finish_reason(mock_response) == "stop"

    def test_response_with_choices_no_finish_reason(self):
        mock_choice = Mock()
        del mock_choice.finish_reason

        mock_response = Mock()
        mock_response.choices = [mock_choice]
        assert get_finish_reason(mock_response) is None

    def test_response_without_choices(self):
        mock_response = Mock()
        del mock_response.choices
        assert get_finish_reason(mock_response) is None

    def test_response_choices_is_none(self):
        mock_response = Mock()
        mock_response.choices = None
        assert get_finish_reason(mock_response) is None

    def test_response_choices_is_empty_list(self):
        mock_response = Mock()
        mock_response.choices = []
        assert get_finish_reason(mock_response) is None

    def test_none_response(self):
        assert get_finish_reason(None) is None


class TestGetOperationType:
    def test_embedding_operation(self):
        messages = []
        kwargs = {"litellm_params": {"api_base": "https://api.example.com/embeddings"}}
        assert get_operation_type(messages, kwargs) == "embedding"

    def test_chat_operation_with_messages(self):
        messages = [{"role": "user", "content": "Hello"}]
        kwargs = {}
        assert get_operation_type(messages, kwargs) == "chat"

    def test_completion_operation_no_messages(self):
        messages = []
        kwargs = {}
        assert get_operation_type(messages, kwargs) == "completion"

    def test_completion_operation_none_messages(self):
        messages = None
        kwargs = {}
        assert get_operation_type(messages, kwargs) == "completion"

    def test_embedding_operation_with_messages(self):
        messages = [{"role": "user", "content": "Hello"}]
        kwargs = {"litellm_params": {"api_base": "https://api.example.com/embeddings"}}
        assert get_operation_type(messages, kwargs) == "embedding"

    def test_embedding_operation_partial_url(self):
        messages = []
        kwargs = {"litellm_params": {"api_base": "/embeddings"}}
        assert get_operation_type(messages, kwargs) == "embedding"

    def test_exception_handling(self):
        messages = []
        kwargs = {"invalid": object()}
        assert get_operation_type(messages, kwargs) == "completion"


class TestGetStreamText:
    def test_response_with_delta_content(self):
        mock_delta = Mock()
        mock_delta.content = "streamed text"

        mock_choice = Mock()
        mock_choice.delta = mock_delta
        del mock_choice.text

        mock_response = Mock()
        mock_response.choices = [mock_choice]
        assert get_stream_text(mock_response) == "streamed text"

    def test_response_with_text_attribute(self):
        mock_choice = Mock()
        del mock_choice.delta
        mock_choice.text = "completion text"

        mock_response = Mock()
        mock_response.choices = [mock_choice]
        assert get_stream_text(mock_response) == "completion text"

    def test_response_with_both_delta_and_text(self):
        mock_delta = Mock()
        mock_delta.content = "streamed text"

        mock_choice = Mock()
        mock_choice.delta = mock_delta
        mock_choice.text = "completion text"

        mock_response = Mock()
        mock_response.choices = [mock_choice]
        assert get_stream_text(mock_response) == "streamed text"

    def test_response_with_none_delta_content(self):
        mock_delta = Mock()
        mock_delta.content = None

        mock_choice = Mock()
        mock_choice.delta = mock_delta
        mock_choice.text = "completion text"

        mock_response = Mock()
        mock_response.choices = [mock_choice]
        assert get_stream_text(mock_response) == "completion text"

    def test_response_with_no_choices(self):
        mock_response = Mock()
        mock_response.choices = None
        assert get_stream_text(mock_response) == ""

    def test_response_with_empty_choices(self):
        mock_response = Mock()
        mock_response.choices = []
        assert get_stream_text(mock_response) == ""

    def test_response_with_none_choices(self):
        mock_response = Mock()
        del mock_response.choices
        assert get_stream_text(mock_response) == ""

    def test_none_response(self):
        assert get_stream_text(None) == ""


class TestGetUsage:
    def test_basic_kwargs(self):
        kwargs = {"max_tokens": 100, "temperature": 0.7, "top_p": 0.9, "stream": True}
        assert get_usage(kwargs) == {
            "max_tokens": 100,
            "temperature": 0.7,
            "top_p": 0.9,
            "stream": True,
        }

    def test_optional_params_fallback(self):
        kwargs = {
            "max_tokens": 100,
            "temperature": 0.7,
            "optional_params": {
                "max_tokens": 200,
                "temperature": 0.8,
                "top_p": 0.95,
                "stream": False,
            },
        }

        assert get_usage(kwargs) == {
            "max_tokens": 100,
            "temperature": 0.7,
            "top_p": 0.95,
            "stream": False,
        }

    def test_mixed_kwargs_and_optional_params(self):
        kwargs = {
            "max_tokens": 100,
            "temperature": 0.7,
            "optional_params": {"top_p": 0.95, "stream": False},
        }

        assert get_usage(kwargs) == {
            "max_tokens": 100,
            "temperature": 0.7,
            "top_p": 0.95,
            "stream": False,
        }

    def test_missing_keys(self):
        kwargs = {"max_tokens": 100, "optional_params": {"temperature": 0.8}}

        assert get_usage(kwargs) == {
            "max_tokens": 100,
            "temperature": 0.8,
            "top_p": None,
            "stream": None,
        }

    def test_empty_kwargs(self):
        kwargs = {}

        assert get_usage(kwargs) == {
            "max_tokens": None,
            "temperature": None,
            "top_p": None,
            "stream": None,
        }

    def test_none_optional_params(self):
        kwargs = {"max_tokens": 100, "optional_params": None}
        assert get_usage(kwargs) == {}

    def test_exception_handling(self):
        kwargs = {"invalid": object()}

        assert get_usage(kwargs) == {
            "max_tokens": None,
            "temperature": None,
            "top_p": None,
            "stream": None,
        }


class TestGetUsageFromResponse:
    def test_response_with_usage_object(self):
        mock_usage = Mock()
        mock_usage.prompt_tokens = 100
        mock_usage.completion_tokens = 50
        mock_usage.total_tokens = 150

        mock_response = Mock()
        mock_response.usage = mock_usage
        assert get_usage_from_response(mock_response) == {
            "prompt_tokens": 100,
            "completion_tokens": 50,
            "total_tokens": 150,
        }

    def test_response_with_usage_object_no_total_tokens(self):
        mock_usage = Mock()
        mock_usage.prompt_tokens = 100
        mock_usage.completion_tokens = 50
        del mock_usage.total_tokens

        mock_response = Mock()
        mock_response.usage = mock_usage
        assert get_usage_from_response(mock_response) == {
            "prompt_tokens": 100,
            "completion_tokens": 50,
            "total_tokens": 150,
        }

    def test_response_with_usage_object_partial_tokens(self):
        mock_usage = Mock()
        mock_usage.prompt_tokens = 100
        del mock_usage.completion_tokens
        del mock_usage.total_tokens

        mock_response = Mock()
        mock_response.usage = mock_usage

        assert get_usage_from_response(mock_response) == {
            "prompt_tokens": 100,
            "completion_tokens": None,
            "total_tokens": 100,
        }

    def test_response_with_usage_object_none_tokens(self):
        mock_usage = Mock()
        mock_usage.prompt_tokens = None
        mock_usage.completion_tokens = None
        del mock_usage.total_tokens

        mock_response = Mock()
        mock_response.usage = mock_usage

        assert get_usage_from_response(mock_response) == {
            "prompt_tokens": None,
            "completion_tokens": None,
            "total_tokens": None,
        }

    def test_response_without_usage_attribute(self):
        mock_response = Mock()
        del mock_response.usage

        assert get_usage_from_response(mock_response) == {
            "prompt_tokens": None,
            "completion_tokens": None,
            "total_tokens": None,
        }

    def test_response_with_none_usage(self):
        mock_response = Mock()
        mock_response.usage = None

        assert get_usage_from_response(mock_response) == {
            "prompt_tokens": None,
            "completion_tokens": None,
            "total_tokens": None,
        }

    def test_none_response(self):
        assert get_usage_from_response(None) == {
            "prompt_tokens": None,
            "completion_tokens": None,
            "total_tokens": None,
        }

    def test_exception_handling(self):
        mock_response = Mock()
        mock_response.usage = object()
        assert get_usage_from_response(mock_response) == {
            "prompt_tokens": None,
            "completion_tokens": None,
            "total_tokens": None,
        }


class TestGetTraceMetadata:
    def test_both_trace_id_and_parent_observation_id(self):
        kwargs = {
            "litellm_params": {
                "metadata": {
                    "trace_id": "trace-123",
                    "parent_observation_id": "obs-456",
                }
            }
        }
        assert get_trace_metadata(kwargs) == {
            "parent_observation_id": "obs-456",
            "trace_id": "trace-123",
        }

    def test_only_trace_id(self):
        kwargs = {
            "litellm_params": {
                "metadata": {
                    "trace_id": "trace-123",
                }
            }
        }
        assert get_trace_metadata(kwargs) == {
            "parent_observation_id": None,
            "trace_id": "trace-123",
        }

    def test_only_parent_observation_id(self):
        kwargs = {
            "litellm_params": {
                "metadata": {
                    "parent_observation_id": "obs-456",
                }
            }
        }
        assert get_trace_metadata(kwargs) == {
            "parent_observation_id": "obs-456",
            "trace_id": None,
        }

    def test_empty_metadata(self):
        kwargs = {"litellm_params": {"metadata": {}}}
        assert get_trace_metadata(kwargs) == {
            "parent_observation_id": None,
            "trace_id": None,
        }

    def test_no_metadata_key(self):
        kwargs = {"litellm_params": {}}
        assert get_trace_metadata(kwargs) == {
            "parent_observation_id": None,
            "trace_id": None,
        }

    def test_no_litellm_params(self):
        kwargs = {}
        assert get_trace_metadata(kwargs) == {
            "parent_observation_id": None,
            "trace_id": None,
        }

    def test_none_litellm_params(self):
        kwargs = {"litellm_params": None}
        assert get_trace_metadata(kwargs) == {
            "parent_observation_id": None,
            "trace_id": None,
        }

    def test_none_metadata(self):
        kwargs = {"litellm_params": {"metadata": None}}
        assert get_trace_metadata(kwargs) == {
            "parent_observation_id": None,
            "trace_id": None,
        }

    def test_empty_kwargs(self):
        kwargs = {}
        assert get_trace_metadata(kwargs) == {
            "parent_observation_id": None,
            "trace_id": None,
        }

    def test_none_kwargs(self):
        assert get_trace_metadata(None) == {
            "parent_observation_id": None,
            "trace_id": None,
        }

    def test_metadata_with_extra_keys(self):
        kwargs = {
            "litellm_params": {
                "metadata": {
                    "trace_id": "trace-123",
                    "parent_observation_id": "obs-456",
                    "extra_key": "extra_value",
                    "another_key": "another_value",
                }
            }
        }
        assert get_trace_metadata(kwargs) == {
            "parent_observation_id": "obs-456",
            "trace_id": "trace-123",
        }

    def test_string_values(self):
        kwargs = {
            "litellm_params": {
                "metadata": {
                    "trace_id": "string-trace-id",
                    "parent_observation_id": "string-obs-id",
                }
            }
        }
        assert get_trace_metadata(kwargs) == {
            "parent_observation_id": "string-obs-id",
            "trace_id": "string-trace-id",
        }

    def test_numeric_values(self):
        kwargs = {
            "litellm_params": {
                "metadata": {
                    "trace_id": 12345,
                    "parent_observation_id": 67890,
                }
            }
        }
        assert get_trace_metadata(kwargs) == {
            "parent_observation_id": 67890,
            "trace_id": 12345,
        }

    def test_boolean_values(self):
        kwargs = {
            "litellm_params": {
                "metadata": {
                    "trace_id": True,
                    "parent_observation_id": False,
                }
            }
        }
        assert get_trace_metadata(kwargs) == {
            "parent_observation_id": False,
            "trace_id": True,
        }

    def test_exception_handling_invalid_structure(self):
        kwargs = {"litellm_params": {"metadata": object()}}
        assert get_trace_metadata(kwargs) == {
            "parent_observation_id": None,
            "trace_id": None,
        }

    def test_exception_handling_invalid_kwargs(self):
        kwargs = {"invalid": object()}
        assert get_trace_metadata(kwargs) == {
            "parent_observation_id": None,
            "trace_id": None,
        }


class TestIntegrationScenarios:
    def test_calculate_total_tokens_integration(self):
        mock_usage = Mock()
        mock_usage.prompt_tokens = 100
        mock_usage.completion_tokens = 50
        del mock_usage.total_tokens

        mock_response = Mock()
        mock_response.usage = mock_usage

        assert get_usage_from_response(mock_response)["total_tokens"] == 150

    def test_get_operation_type_with_embedding_url_variations(self):
        test_cases = [
            ("https://api.openai.com/v1/embeddings", "embedding"),
            ("http://localhost:8000/embeddings", "embedding"),
            ("/embeddings", "embedding"),
            ("embeddings", "completion"),
            ("https://api.openai.com/v1/chat/completions", "completion"),
            ("", "completion"),
        ]

        for api_base, expected in test_cases:
            kwargs = {"litellm_params": {"api_base": api_base}}
            result = get_operation_type([], kwargs)
            assert result == expected, f"Failed for api_base: {api_base}"

    def test_get_stream_text_with_various_chunk_formats(self):
        mock_delta = Mock()
        mock_delta.content = "Hello"

        mock_choice = Mock()
        mock_choice.delta = mock_delta
        del mock_choice.text

        mock_response = Mock()
        mock_response.choices = [mock_choice]

        result = get_stream_text(mock_response)
        assert result == "Hello"

        mock_choice2 = Mock()
        del mock_choice2.delta
        mock_choice2.text = "World"

        mock_response2 = Mock()
        mock_response2.choices = [mock_choice2]

        result2 = get_stream_text(mock_response2)
        assert result2 == "World"



import logging
from collections.abc import Sequence
from pathlib import Path

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult

logger = logging.getLogger(__name__)


class FileSpanExporter(SpanExporter):
    """Span exporter that writes each span to its own JSON file in a folder."""

    def __init__(self, folder_path: str):
        self.folder_path = Path(folder_path)
        self.folder_path.mkdir(parents=True, exist_ok=True)

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        """Export each span to its own file."""
        try:
            for span in spans:
                trace_id = format(span.context.trace_id, "032x")
                span_id = format(span.context.span_id, "016x")
                filename = f"{trace_id}_{span_id}.json"
                file_path = self.folder_path / filename

                with open(file_path, "w", encoding="utf-8") as f:
                    f.write(span.to_json())

            return SpanExportResult.SUCCESS
        except Exception:
            return SpanExportResult.FAILURE

    def shutdown(self) -> None:
        """Shutdown the exporter."""



from .api import (
    CHARACTER_DETAILS_QUESTION,
    FRIENDS_QUESTION,
    STORY_QUESTION,
    get_all_characters,
    get_all_episodes,
    get_all_items,
    get_all_locations,
    get_all_occupations,
    get_character_by_id,
    get_character_location,
    get_character_occupation,
    get_characters_by_episode,
    get_characters_by_item,
    get_characters_by_location,
    get_characters_by_occupation,
    get_episode_by_id,
    get_episodes_by_character,
    get_episodes_by_item,
    get_episodes_by_location,
    get_favorite_items,
    get_friends,
    get_item_by_id,
    get_items_by_episode,
    get_location_by_episode,
    get_location_by_id,
    get_occupation_by_id,
)

datasource = [
    get_all_characters,
    get_all_episodes,
    get_all_items,
    get_all_locations,
    get_all_occupations,
    get_character_by_id,
    get_character_location,
    get_character_occupation,
    get_characters_by_episode,
    get_characters_by_item,
    get_characters_by_location,
    get_characters_by_occupation,
    get_episode_by_id,
    get_episodes_by_character,
    get_episodes_by_item,
    get_episodes_by_location,
    get_favorite_items,
    get_friends,
    get_item_by_id,
    get_items_by_episode,
    get_location_by_episode,
    get_location_by_id,
    get_occupation_by_id,
]


__all__ = [
    CHARACTER_DETAILS_QUESTION,
    FRIENDS_QUESTION,
    STORY_QUESTION,
    datasource,
]



import logging
from typing import Any

from ...contexts import CompletionChoice, GenerationUsage


def get_finish_reason(response_obj: Any) -> str | None:
    """Extract finish reason from Google GenAI response."""
    try:
        if hasattr(response_obj, "candidates") and response_obj.candidates:
            candidate = response_obj.candidates[0]
            if hasattr(candidate, "finish_reason"):
                finish_reason = str(candidate.finish_reason)
                if "STOP" in finish_reason:
                    return "stop"
                elif "MAX_TOKENS" in finish_reason:
                    return "length"
                elif "SAFETY" in finish_reason or "RECITATION" in finish_reason:
                    return "content_filter"
                else:
                    return "other"
        return None
    except Exception as e:
        logging.warning("Error getting finish reason from Google GenAI", exc_info=e)
        return None


def get_completion_choices(response_obj: Any) -> list[CompletionChoice]:
    """Extract completion choices from Google GenAI response."""
    try:
        choices = []
        if hasattr(response_obj, "candidates") and response_obj.candidates:
            for candidate in response_obj.candidates:
                if hasattr(candidate, "content") and candidate.content:
                    content = candidate.content

                    text_parts = []
                    tool_calls = []

                    if hasattr(content, "parts") and content.parts:
                        for part in content.parts:
                            if hasattr(part, "text") and part.text:
                                text_parts.append(part.text)
                            elif hasattr(part, "function_call"):
                                func_call = part.function_call
                                tool_call = {
                                    "id": getattr(func_call, "id", f"call_{hash(func_call.name)}"),
                                    "type": "function",
                                    "function": {
                                        "name": func_call.name,
                                        "arguments": str(func_call.args),
                                    },
                                }
                                tool_calls.append(tool_call)

                    text_content = "".join(text_parts) if text_parts else None

                    choice = CompletionChoice(
                        message={
                            "role": getattr(content, "role", "model"),
                            "content": text_content,
                            "function_call": None,
                            "name": None,
                            "tool_calls": tool_calls if tool_calls else None,
                        },
                        delta=None,
                        text=text_content,
                    )
                    choices.append(choice)

        return choices
    except Exception as e:
        logging.warning("Error getting completion choices from Google GenAI", exc_info=e)
        return []


def get_usage_from_response(response_obj: Any) -> GenerationUsage:
    """Extract usage information from Google GenAI response."""
    try:
        if hasattr(response_obj, "usage_metadata"):
            usage = response_obj.usage_metadata
            prompt_tokens = getattr(usage, "prompt_token_count", None)
            completion_tokens = getattr(usage, "response_token_count", None)
            total_tokens = getattr(usage, "total_token_count", None)

            return {
                "input_tokens": prompt_tokens,
                "output_tokens": completion_tokens,
                "prompt_tokens": prompt_tokens,
                "completion_tokens": completion_tokens,
                "total_tokens": total_tokens,
                "max_tokens": None,
                "temperature": None,
                "top_p": None,
                "stream": None,
                "tool_calls": None,
            }
        return {}
    except Exception as e:
        logging.warning("Error getting usage from Google GenAI response", exc_info=e)
        return {}


def get_request_parameters(kwargs: dict) -> GenerationUsage:
    """Extract request parameters from kwargs."""
    try:
        config = kwargs.get("config")
        max_tokens = None
        temperature = None
        top_p = None

        if config:
            if isinstance(config, dict):
                max_tokens = config.get("max_output_tokens")
                temperature = config.get("temperature")
                top_p = config.get("top_p")
            else:
                if hasattr(config, "max_output_tokens"):
                    max_tokens = config.max_output_tokens
                if hasattr(config, "temperature"):
                    temperature = config.temperature
                if hasattr(config, "top_p"):
                    top_p = config.top_p

        return {
            "max_tokens": max_tokens,
            "temperature": temperature,
            "top_p": top_p,
            "stream": False,
            "input_tokens": None,
            "output_tokens": None,
            "prompt_tokens": None,
            "completion_tokens": None,
            "total_tokens": None,
            "tool_calls": None,
        }
    except Exception as e:
        logging.warning("Error getting request parameters from Google GenAI", exc_info=e)
        return {}


def format_messages_for_tracing(contents: Any) -> list[dict]:
    """Format Google GenAI contents for tracing."""
    try:
        formatted_messages = []

        if isinstance(contents, str):
            formatted_messages.append({
                "role": "user",
                "content": contents,
                "function_call": None,
                "name": None,
                "tool_calls": None,
            })
        elif isinstance(contents, list):
            for item in contents:
                if isinstance(item, str):
                    formatted_messages.append({
                        "role": "user",
                        "content": item,
                        "function_call": None,
                        "name": None,
                        "tool_calls": None,
                    })
                elif isinstance(item, dict):
                    role = item.get("role", "user")
                    parts = item.get("parts", [])

                    text_content = []
                    for part in parts:
                        if isinstance(part, str):
                            text_content.append(part)
                        elif isinstance(part, dict):
                            if "text" in part:
                                text_content.append(part["text"])
                            elif "function_call" in part:
                                text_content.append(f"[Function Call: {part['function_call'].get('name', 'unknown')}]")
                            elif "function_response" in part:
                                text_content.append("[Function Response]")

                    formatted_messages.append({
                        "role": role,
                        "content": "\n".join(text_content) if text_content else "",
                        "function_call": None,
                        "name": None,
                        "tool_calls": None,
                    })
                elif hasattr(item, "role") and hasattr(item, "parts"):
                    role = getattr(item, "role", "user")
                    parts = getattr(item, "parts", [])

                    text_content = []
                    for part in parts:
                        if hasattr(part, "text"):
                            text_content.append(part.text)
                        elif hasattr(part, "function_call"):
                            func_call = part.function_call
                            text_content.append(f"[Function Call: {func_call.name}]")
                        elif hasattr(part, "function_response"):
                            text_content.append("[Function Response]")

                    formatted_messages.append({
                        "role": role,
                        "content": "\n".join(text_content) if text_content else "",
                        "function_call": None,
                        "name": None,
                        "tool_calls": None,
                    })
        elif hasattr(contents, "role") and hasattr(contents, "parts"):
            role = getattr(contents, "role", "user")
            parts = getattr(contents, "parts", [])

            text_content = []
            for part in parts:
                if hasattr(part, "text"):
                    text_content.append(part.text)
                elif hasattr(part, "function_call"):
                    func_call = part.function_call
                    text_content.append(f"[Function Call: {func_call.name}]")
                elif hasattr(part, "function_response"):
                    text_content.append("[Function Response]")

            formatted_messages.append({
                "role": role,
                "content": "\n".join(text_content) if text_content else "",
                "function_call": None,
                "name": None,
                "tool_calls": None,
            })

        return formatted_messages
    except Exception as e:
        logging.warning("Error formatting messages for tracing", exc_info=e)
        return []


def get_stream_text(chunk: Any) -> str:
    """Extract text from streaming chunk."""
    try:
        if hasattr(chunk, "text") and chunk.text:
            return chunk.text

        if hasattr(chunk, "candidates") and chunk.candidates:
            candidate = chunk.candidates[0]
            if hasattr(candidate, "content") and candidate.content:
                content = candidate.content
                if hasattr(content, "parts") and content.parts:
                    text_parts = []
                    for part in content.parts:
                        if hasattr(part, "text") and part.text:
                            text_parts.append(part.text)
                    return "".join(text_parts)

        return ""
    except Exception:
        return ""



import importlib
import logging
from collections.abc import Callable
from typing import Literal, TypedDict

from ..sdk import VerseSDK

ANTHROPIC_MODULE = "anthropic"
LANGCHAIN_MODULE = "langchain"
LITELLM_MODULE = "litellm"
GOOGLE_MODULE = "google"
PYDANTIC_AI_MODULE = "pydantic_ai"

VENDOR_ALLOWLIST = Literal[
    ANTHROPIC_MODULE,
    LANGCHAIN_MODULE,
    LITELLM_MODULE,
    GOOGLE_MODULE,
    PYDANTIC_AI_MODULE,
]

VENDOR_ALLOWLIST_CONFIG = list[VENDOR_ALLOWLIST] | VENDOR_ALLOWLIST | None


class VendorIntegration(TypedDict):
    """Configuration for a vendor integration."""

    module_name: str
    configure_func: str
    cleanup_func: str | None


INTEGRATIONS: dict[str, VendorIntegration] = {
    ANTHROPIC_MODULE: {
        "module_name": ANTHROPIC_MODULE,
        "configure_func": "configure_anthropic",
        "cleanup_func": None,
    },
    LANGCHAIN_MODULE: {
        "module_name": LANGCHAIN_MODULE,
        "configure_func": "configure_langchain",
        "cleanup_func": None,
    },
    LITELLM_MODULE: {
        "module_name": LITELLM_MODULE,
        "configure_func": "configure_litellm",
        "cleanup_func": "cleanup_litellm",
    },
    GOOGLE_MODULE: {
        "module_name": GOOGLE_MODULE,
        "configure_func": "configure_google",
        "cleanup_func": None,
    },
    PYDANTIC_AI_MODULE: {
        "module_name": PYDANTIC_AI_MODULE,
        "configure_func": "patch_pydantic_ai",
        "cleanup_func": None,
    },
}


def _load_integration(integration: VendorIntegration, sdk: VerseSDK) -> Callable | None:
    """
    Load and apply a vendor integration.

    Returns the cleanup function if one exists, otherwise None.
    """
    # Use importlib.import_module instead of __import__ for better relative import handling
    module_name = f"verse_sdk.integrations.{integration['module_name']}"
    module = importlib.import_module(module_name)

    getattr(module, integration["configure_func"])(sdk)

    if integration["cleanup_func"]:
        return getattr(module, integration["cleanup_func"])
    return None


def apply_integrations(
    sdk: VerseSDK,
    vendors: VENDOR_ALLOWLIST_CONFIG,
) -> Callable:
    """
    Apply vendor integrations to the VerseSDK.

    Args:
        sdk: The VerseSDK instance to configure
        vendors: Vendor name(s) to integrate, or None for no integrations

    Returns:
        A cleanup function that should be called to teardown integrations
    """
    cleanup_functions = []

    def _cleanup_integrations():
        """Cleanup all applied integrations."""
        for cleanup_fn in cleanup_functions:
            try:
                cleanup_fn()
            except Exception as e:
                logging.warning(
                    f"Error cleaning up `{cleanup_fn.__name__}`", exc_info=e
                )

    if not vendors:
        return _cleanup_integrations

    vendor_list = [vendors] if isinstance(vendors, str) else vendors

    for vendor in vendor_list:
        vendor = vendor.lower()

        integration = INTEGRATIONS.get(vendor)
        if not integration:
            logging.warning(f"Unknown vendor integration: `{vendor}`")
            continue

        try:
            cleanup_fn = _load_integration(integration, sdk)
            if cleanup_fn:
                cleanup_functions.append(cleanup_fn)
            logging.info(f"`{vendor}` integration applied to `VerseSDK`")
        except Exception as e:
            logging.warning(f"Error applying `{vendor}` integration", exc_info=e)

    return _cleanup_integrations


__all__ = [
    "VENDOR_ALLOWLIST_CONFIG",
    "apply_integrations",
]



from unittest.mock import Mock, patch

import pytest

from verse_sdk.sdk import VerseSDK


class TestVerseSDK:
    def setup_method(self):
        self.mock_tracer = Mock()
        self.mock_span = Mock()

        self.mock_trace_patcher = patch("verse_sdk.sdk.trace")
        self.mock_trace = self.mock_trace_patcher.start()
        self.mock_trace.get_current_span.return_value = self.mock_span
        self.mock_provider = Mock()
        self.mock_trace.get_tracer_provider.return_value = self.mock_provider

        class MockNoOpTracerProvider:
            pass

        self.mock_trace.NoOpTracerProvider = MockNoOpTracerProvider

    def teardown_method(self):
        self.mock_trace_patcher.stop()

    def test_score_with_observation_target(self):
        sdk = VerseSDK(self.mock_tracer)
        score_data = {"accuracy": 0.95, "confidence": 0.8}

        with patch("verse_sdk.sdk.SpanContext") as mock_span_context:
            mock_span_instance = Mock()
            mock_span_context.return_value = mock_span_instance

            sdk.score(score_data, "observation")

            mock_span_context.assert_called_once_with(self.mock_span)
            mock_span_instance.score.assert_called_once_with(score_data)

    def test_score_with_trace_target(self):
        sdk = VerseSDK(self.mock_tracer)
        score_data = {"overall": 0.9, "quality": 0.85}

        with patch("verse_sdk.sdk.TraceContext") as mock_trace_context:
            mock_trace_instance = Mock()
            mock_trace_context.return_value = mock_trace_instance

            sdk.score(score_data, "trace")

            mock_trace_context.assert_called_once_with(self.mock_span)
            mock_trace_instance.score.assert_called_once_with(score_data)

    def test_score_handles_exception(self):
        sdk = VerseSDK(self.mock_tracer)
        score_data = {"test": "value"}

        with patch("verse_sdk.sdk.SpanContext") as mock_span_context:
            mock_span_context.side_effect = Exception("Test error")

            with patch("verse_sdk.sdk.logging") as mock_logging:
                sdk.score(score_data, "observation")

                mock_logging.warning.assert_called_once()

    def test_set_tracer_with_valid_tracer(self):
        sdk = VerseSDK(self.mock_tracer)
        assert sdk._tracer == self.mock_tracer

    def test_set_tracer_with_none_tracer(self):
        with pytest.raises(ValueError, match="Tracer not initialized"):
            VerseSDK(None)

    def test_set_tracer_with_false_tracer(self):
        with pytest.raises(ValueError, match="Tracer not initialized"):
            VerseSDK(False)

    def test_set_tracer_with_empty_string_tracer(self):
        with pytest.raises(ValueError, match="Tracer not initialized"):
            VerseSDK("")

    def test_set_tracer_with_zero_tracer(self):
        with pytest.raises(ValueError, match="Tracer not initialized"):
            VerseSDK(0)

    def test_validate_tracer_provider_with_valid_provider(self):
        sdk = VerseSDK(self.mock_tracer)
        sdk._validate_tracer_provider()

    def test_validate_tracer_provider_with_exception(self):
        self.mock_trace.get_tracer_provider.side_effect = Exception("Provider error")

        with pytest.raises(ValueError, match="Tracer provider unavailable"):
            VerseSDK(self.mock_tracer)

    def test_validate_tracer_provider_with_none_provider(self):
        self.mock_trace.get_tracer_provider.return_value = None

        with pytest.raises(ValueError, match="Invalid tracer provider given"):
            VerseSDK(self.mock_tracer)

    def test_validate_tracer_provider_with_noop_provider(self):
        mock_noop_provider = Mock()
        mock_noop_provider.__class__ = self.mock_trace.NoOpTracerProvider
        self.mock_trace.get_tracer_provider.return_value = mock_noop_provider

        with pytest.raises(ValueError, match="Invalid tracer provider given"):
            VerseSDK(self.mock_tracer)

    def test_score_with_different_score_types(self):
        sdk = VerseSDK(self.mock_tracer)

        with patch("verse_sdk.sdk.SpanContext") as mock_span_context:
            mock_span_instance = Mock()
            mock_span_context.return_value = mock_span_instance

            numeric_score = {"rating": 4.5}
            sdk.score(numeric_score, "observation")
            mock_span_instance.score.assert_called_with(numeric_score)

            boolean_score = {"passed": True, "failed": False}
            sdk.score(boolean_score, "observation")
            mock_span_instance.score.assert_called_with(boolean_score)

            string_score = {"status": "excellent", "grade": "A"}
            sdk.score(string_score, "observation")
            mock_span_instance.score.assert_called_with(string_score)

    def test_score_with_empty_score(self):
        sdk = VerseSDK(self.mock_tracer)

        with patch("verse_sdk.sdk.SpanContext") as mock_span_context:
            mock_span_instance = Mock()
            mock_span_context.return_value = mock_span_instance

            empty_score = {}
            sdk.score(empty_score, "observation")
            mock_span_instance.score.assert_called_once_with(empty_score)

    def test_score_with_nested_score(self):
        sdk = VerseSDK(self.mock_tracer)

        with patch("verse_sdk.sdk.SpanContext") as mock_span_context:
            mock_span_instance = Mock()
            mock_span_context.return_value = mock_span_instance

            nested_score = {
                "metrics": {"accuracy": 0.95, "precision": 0.92, "recall": 0.88},
                "overall": 0.91,
            }
            sdk.score(nested_score, "observation")
            mock_span_instance.score.assert_called_once_with(nested_score)



from __future__ import annotations

from typing import Any

from ..utils import to_json
from .base import BaseContext, ContextType
from .types import EventMetadata, Level, ObservationType


class SpanContext(BaseContext):
    """Context for regular spans (sub-operations)."""

    observation_type: ObservationType = "observation"
    type: ContextType = "span"

    def event(
        self,
        name: str,
        *,
        level: Level = "info",
        metadata: EventMetadata | None = None,
        **attrs,
    ) -> SpanContext:
        """Add an event to the span with severity level and metadata."""
        if self._span.get_span_context().is_valid:
            payload = {"level": level}
            if metadata:
                payload.update(metadata=to_json(metadata))
            payload.update(**attrs)
            self._span.add_event(name, payload)
        return self

    def input(self, input: Any) -> SpanContext:
        """Set the input for the span."""
        self._span.set_attribute("input", str(input))
        return self

    def level(self, level: Level) -> SpanContext:
        """Set the level for the span."""
        self._span.set_attribute("observation.level", str(level))
        return self

    def output(self, output: Any) -> SpanContext:
        """Set the output for the span."""
        self._span.set_attribute("output", str(output))
        return self

    def status_message(self, status_message: str) -> SpanContext:
        """Set the status message for the span."""
        self._span.set_attribute("observation.status_message", str(status_message))
        return self



import logging
from typing import Any

from opentelemetry.trace import Span

from ...contexts import CompletionChoice, GenerationUsage, create_span_in_existing_trace
from ...utils import uses_pydantic_base_model


def calculate_total_tokens(
    prompt_tokens: int | None, completion_tokens: int | None
) -> int | None:
    """Calculate the total tokens."""
    if prompt_tokens is not None or completion_tokens is not None:
        return (prompt_tokens or 0) + (completion_tokens or 0)
    return None


def get_completition_choices(data: Any) -> list[CompletionChoice]:
    """Get the completion choices from the response."""

    choices = []
    if hasattr(data, "choices"):
        choices = data.choices
    elif isinstance(data, dict) and "choices" in data:
        choices = data["choices"]

    converted_choices = []
    for choice in choices:
        if uses_pydantic_base_model(choice):
            converted_choices.append(choice.model_dump())
        elif isinstance(choice, dict):
            converted_choices.append(choice)
        else:
            # Fallback: try to convert to dict
            converted_choices.append(
                dict(choice) if hasattr(choice, "__dict__") else choice
            )

    return converted_choices if len(converted_choices) > 0 else []


def get_finish_reason(response_obj: Any) -> str | None:
    """Trace response metadata like model and ID."""
    if getattr(response_obj, "choices", None):
        fr = getattr(response_obj.choices[0], "finish_reason", None)
        if fr:
            return fr

    return None


def get_operation_type(messages: list, kwargs: dict) -> str:
    """Determine the operation type based on request parameters."""
    try:
        if kwargs.get("litellm_params", {}).get("api_base", "").endswith("/embeddings"):
            return "embedding"

        return "chat" if messages else "completion"
    except Exception as e:
        logging.warning("Error determining operation type from Litellm", exc_info=e)
        return "completion"


def get_stream_text(response_obj: Any) -> str:
    """Get the text from the response."""
    chunk = getattr(response_obj, "choices", None)
    if not chunk:
        return ""

    c = chunk[0]
    return (
        getattr(getattr(c, "delta", None), "content", None)
        or getattr(c, "text", None)
        or ""
    )


def get_trace_metadata(kwargs: dict) -> dict[str, str | None]:
    """
    Extract trace ID and parent observation ID from LiteLLM metadata.
    Useful when using Router().completion() or parallel processing.
    """
    try:
        litellm_params = kwargs.get("litellm_params", {}).get("metadata", {})
        parent_observation_id = litellm_params.get("parent_observation_id", None)
        trace_id = litellm_params.get("trace_id", None)
        return {
            "parent_observation_id": parent_observation_id,
            "trace_id": trace_id,
        }
    except Exception:
        return {
            "parent_observation_id": None,
            "trace_id": None,
        }


def get_trace_span_from_metadata(kwargs: dict) -> Span | None:
    """
    Generate parent trace from metadata.
    Needed when the LLM call is outside of the context manager.
    """
    try:
        trace_metadata = get_trace_metadata(kwargs)
        if (
            trace_metadata
            and trace_metadata.get("trace_id")
            and trace_metadata.get("parent_observation_id")
        ):
            return create_span_in_existing_trace(
                trace_metadata.get("trace_id"),
                trace_metadata.get("parent_observation_id"),
            )
    except Exception:
        return None


def get_usage(kwargs: dict) -> GenerationUsage:
    """Get the usage for the request."""
    try:
        optional = kwargs.get("optional_params", {})
        return {
            "max_tokens": kwargs.get("max_tokens", optional.get("max_tokens")),
            "temperature": kwargs.get("temperature", optional.get("temperature")),
            "top_p": kwargs.get("top_p", optional.get("top_p")),
            "stream": kwargs.get("stream", optional.get("stream")),
        }
    except Exception as e:
        logging.warning("Error getting usage from Litellm", exc_info=e)
        return {}


def get_usage_from_response(response_obj: Any) -> GenerationUsage:
    """Get the usage from the response."""
    try:
        usage = getattr(response_obj, "usage", None)

        usage_output = {
            "prompt_tokens": getattr(usage, "prompt_tokens", None),
            "completion_tokens": getattr(usage, "completion_tokens", None),
        }

        usage_output["total_tokens"] = getattr(
            usage,
            "total_tokens",
            calculate_total_tokens(
                usage_output["prompt_tokens"],
                usage_output["completion_tokens"],
            ),
        )

        return usage_output

    except Exception as e:
        logging.warning("Error getting usage from response", exc_info=e)
        return {}



import os

from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter as HTTPTraceExporter,
)
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import SpanProcessor

from ...processors import AttributeBatchSpanProcessor
from ..base import Exporter
from .constants import VERSE_API_KEY, VERSE_HOST, VERSE_PROJECT_ID
from .types import VerseConfig
from .utils import as_traces_endpoint


class VerseExporter(Exporter):
    """Verse HTTP trace exporter."""

    def __init__(self, config: VerseConfig = None):
        self.config = config or VerseExporter.env()

    @property
    def endpoint(self) -> str:
        """Get the Verse HTTP endpoint."""
        return as_traces_endpoint(
            self.config.get("host"),
            self.config.get("api_key"),
        )

    def create_span_processor(self, resource: Resource) -> SpanProcessor:
        if not self.config.get("project_id"):
            raise ValueError("Project ID is required to export traces to Verse")

        return AttributeBatchSpanProcessor(
            HTTPTraceExporter(self.endpoint),
            self.config.get("project_id"),
        )

    def get_name(self) -> str:
        return "verse"

    @staticmethod
    def env() -> VerseConfig:
        """Get the Verse config from environment variables."""
        return VerseConfig(
            api_key=os.environ.get(VERSE_API_KEY),
            host=os.environ.get(VERSE_HOST),
            project_id=os.environ.get(VERSE_PROJECT_ID),
        )



import os

from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter as HTTPTraceExporter,
)
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import SpanProcessor
from opentelemetry.sdk.trace.export import BatchSpanProcessor

from ..base import Exporter
from ..verse.utils import as_traces_endpoint
from .constants import OTEL_HOST
from .types import OtelConfig


class OTLPExporter(Exporter):
    """Generic OTLP exporter for OpenTelemetry-compatible backends and collectors."""

    def __init__(self, config: OtelConfig = None):
        self.config = config or OTLPExporter.env()

    @property
    def endpoint(self) -> str:
        """Get the OTel HTTP endpoint."""
        return as_traces_endpoint(self.config.get("host"))

    def create_span_processor(self, resource: Resource) -> SpanProcessor:
        exporter = HTTPTraceExporter(self.endpoint)
        return BatchSpanProcessor(exporter)

    def get_name(self) -> str:
        return "otlp"

    @staticmethod
    def env() -> OtelConfig:
        """Get the langfuse config from environment variables."""
        return OtelConfig(
            host=os.environ.get(OTEL_HOST),
        )



from typing import ClassVar

from opentelemetry.context import Context
from opentelemetry.sdk.trace import ReadableSpan, Span, SpanProcessor


class AttributeProcessor(SpanProcessor):
    _project_id: str
    _propagated_attributes: ClassVar[list[str]] = [
        "session.id",
        "project.id",
        "user.id",
    ]

    def __init__(self, project_id: str):
        """Initialize the processor with span tracking structures."""
        self._active_spans: dict[int, Span] = {}
        self._project_id = project_id
        self._span_cache: dict[int, dict[str, str]] = {}

    def force_flush(self, timeout_millis: int = 30000):
        """Force flush any buffered spans. This processor has no buffer."""
        return True

    def on_end(self, span: ReadableSpan):
        """Called when a span ends. Cache attributes for future children and cleanup."""
        span_id = span.get_span_context().span_id

        if span.attributes:
            attribute_store = {
                attribute: span.attributes[attribute]
                for attribute in self._propagated_attributes
                if attribute in span.attributes
            }
            if attribute_store:
                self._span_cache[span_id] = attribute_store

        self._active_spans.pop(span_id, None)

    def on_start(self, span: Span, parent_context: Context | None = None):
        """Called when a span starts. Propagates attributes from parent and registers span."""
        span_id = span.get_span_context().span_id
        self._active_spans[span_id] = span
        self._inject_project_id(span)

        if not span.parent or not span.parent.is_valid:
            return

        self._propagate_attributes(span, span.parent.span_id)

    def shutdown(self):
        """Called when the tracer provider shuts down. Cleans up all caches."""
        self._active_spans.clear()
        self._span_cache.clear()
        return True

    def _copy_attributes(self, source_attributes: dict, target_span: Span):
        """Copy propagated attributes from parent to child span."""
        for attribute in self._propagated_attributes:
            attribute_value = source_attributes.get(attribute)
            if attribute_value is not None:
                target_span.set_attribute(attribute, attribute_value)

    def _inject_project_id(self, span: Span):
        if self._project_id:
            span.set_attribute("project.id", self._project_id)

    def _propagate_attributes(self, span: Span, parent_span_id: int):
        """Propagate configured attributes from parent to child span."""
        parent_span = self._active_spans.get(parent_span_id)
        if parent_span and hasattr(parent_span, "_attributes"):
            self._copy_attributes(parent_span._attributes, span)
            return

        cached_attributes = self._span_cache.get(parent_span_id)
        if cached_attributes:
            self._copy_attributes(cached_attributes, span)



from __future__ import annotations

from typing import Any

from .base import BaseContext
from .types import ContextType, ObservationType


class TraceContext(BaseContext):
    """Context for trace-level spans (top-level operations)."""

    observation_type: ObservationType = "trace"
    type: ContextType = "trace"

    def input(self, input: Any) -> TraceContext:
        """Set the input for the trace."""
        self._span.set_attribute("input", str(input))
        return self

    def project(self, project_id: str) -> TraceContext:
        """Set the project ID for the trace."""
        self._span.set_attribute("project.id", project_id)
        return self

    def session(self, session_id: str) -> TraceContext:
        """Set the session ID for the trace."""
        self._span.set_attribute("session.id", session_id)
        return self

    def output(self, output: Any) -> TraceContext:
        """Set the output for the trace."""
        self._span.set_attribute("output", str(output))
        return self

    def scope(self, scope: str) -> TraceContext:
        """Set the scope for the trace."""
        self._span.set_attribute("session.scope", scope)
        return self

    def tags(self, tags: list[str]) -> TraceContext:
        """Add tags to the trace."""
        self._span.set_attribute("tags", tags)
        return self

    def update(self, **kwargs) -> TraceContext:
        """Update the trace with additional attributes."""
        self.set_attributes(**kwargs)
        return self

    def user(self, user_id: str) -> TraceContext:
        """Set the user ID for the trace."""
        self._span.set_attribute("user.id", user_id)
        return self



from opentelemetry.sdk.trace.export import SpanExporter

from ..exporters.types import ExporterConfig


class ScopeFilterExporter(SpanExporter):
    _allowed_scopes: list[str]
    _exporter: SpanExporter

    def __init__(self, exporter: SpanExporter, config: ExporterConfig):
        self._allowed_scopes = config.get("scopes", [])
        self._exporter = exporter

    def export(self, spans):
        """Export the spans, filtering out anything that doesn't have an allowed scope."""
        if not self.is_enabled():
            return self._exporter.export(spans)

        exported_spans = []
        trace_scopes = {}

        for span in spans:
            scope = span.attributes.get("session.scope")
            trace_id = span.get_span_context().trace_id

            if scope is not None and trace_id not in trace_scopes:
                trace_scopes[trace_id] = scope

            # assumes that trace wraps fully observation
            if self._has_scope(trace_scopes.get(trace_id)):
                exported_spans.append(span)

        return self._exporter.export(exported_spans)

    def force_flush(self, timeout_millis=None):
        """Force flush the scope filter processor."""
        return self._exporter.force_flush(timeout_millis)

    def shutdown(self):
        """Shutdown the scope filter processor."""
        return self._exporter.shutdown()

    def _has_scope(self, scope: str | None) -> bool:
        """Check if the span has a scope."""
        return scope and scope in self._allowed_scopes

    def is_enabled(self) -> bool:
        """Check if the scope filter processor is enabled."""
        return self._allowed_scopes is not None and len(self._allowed_scopes) > 0



from unittest.mock import patch

import pytest
from opentelemetry import trace as otel_trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult

from verse_sdk import VerseConfig, verse


class InMemorySpanExporter(SpanExporter):
    def __init__(self):
        self.spans = []

    def export(self, spans):
        self.spans.extend(spans)
        return SpanExportResult.SUCCESS

    def shutdown(self):
        pass

    def get_finished_spans(self):
        return self.spans


def get_spans_from_exporter(exporter: InMemorySpanExporter) -> list[dict]:
    spans = []
    for span in exporter.get_finished_spans():
        span_dict = {
            "name": span.name,
            "attributes": dict(span.attributes) if span.attributes else {},
            "context": {
                "trace_id": format(span.context.trace_id, "032x"),
                "span_id": format(span.context.span_id, "016x"),
            },
        }
        spans.append(span_dict)
    return spans


def force_flush():
    provider = otel_trace.get_tracer_provider()
    if isinstance(provider, TracerProvider):
        provider.force_flush()


@pytest.fixture(autouse=True)
def reset_sdk():
    yield
    verse._initialized = False
    verse._sdk = None
    verse._tracer = None
    provider = otel_trace.get_tracer_provider()
    if isinstance(provider, TracerProvider):
        provider.shutdown()
    otel_trace._TRACER_PROVIDER = None
    otel_trace._TRACER_PROVIDER_SET_ONCE._done = False


@patch("verse_sdk.exporters.verse.exporter.HTTPTraceExporter")
def test_context_managers(mock_http_exporter_class):
    memory_exporter = InMemorySpanExporter()
    mock_http_exporter_class.return_value = memory_exporter

    verse.init(
        app_name="test_context_managers",
        exporters=[
            verse.exporters.verse(
                VerseConfig(
                    project_id="test-project",
                    api_key="test-api-key",
                    host="test-host",
                )
            )
        ],
    )

    with verse.trace(name="test_trace") as trace:
        trace.input("test_input")
        trace.output("test_output")

        with verse.span(name="test_span") as span:
            span.input("test_span_input")
            span.output("test_span_output")

            with verse.generation(name="test_generation") as generation:
                generation.model("test_model")
                generation.input("test_generation_input")
                generation.output("test_generation_output")

    force_flush()

    spans = get_spans_from_exporter(memory_exporter)
    assert len(spans) == 3

    assert spans[0]["name"] == "test_generation"
    assert spans[0]["attributes"]["gen_ai.request.model"] == "test_model"
    assert spans[0]["attributes"]["gen_ai.prompt"] == "test_generation_input"
    trace_id = spans[0]["context"]["trace_id"]

    assert spans[1]["name"] == "test_span"
    assert spans[1]["context"]["trace_id"] == trace_id
    assert spans[1]["attributes"]["input"] == "test_span_input"
    assert spans[1]["attributes"]["output"] == "test_span_output"

    assert spans[2]["name"] == "test_trace"
    assert spans[2]["context"]["trace_id"] == trace_id
    assert spans[2]["attributes"]["input"] == "test_input"
    assert spans[2]["attributes"]["output"] == "test_output"


@patch("verse_sdk.exporters.verse.exporter.HTTPTraceExporter")
def test_session_and_project_propagation(mock_http_exporter_class):
    memory_exporter = InMemorySpanExporter()
    mock_http_exporter_class.return_value = memory_exporter

    verse.init(
        app_name="test_context_managers",
        exporters=[
            verse.exporters.verse(
                VerseConfig(
                    api_key="test-api-key",
                    host="test-host",
                    project_id="test-project",
                )
            )
        ],
    )

    with verse.trace(
        name="parent_trace",
        session_id="test-session-123",
    ) as trace:
        trace.input("parent_input")

        with verse.span(name="child_span") as span:
            span.input("child_span_input")

            with verse.generation(name="child_generation") as generation:
                generation.model("test_model")
                generation.input("child_generation_input")

    force_flush()

    spans = get_spans_from_exporter(memory_exporter)
    assert len(spans) == 3

    assert spans[0]["name"] == "child_generation"
    assert spans[0]["attributes"]["session.id"] == "test-session-123"
    assert spans[0]["attributes"]["project.id"] == "test-project"

    assert spans[1]["name"] == "child_span"
    assert spans[1]["attributes"]["session.id"] == "test-session-123"
    assert spans[1]["attributes"]["project.id"] == "test-project"

    assert spans[2]["name"] == "parent_trace"
    assert spans[2]["attributes"]["session.id"] == "test-session-123"
    assert spans[2]["attributes"]["project.id"] == "test-project"



from unittest.mock import Mock

from verse_sdk.integrations.anthropic.utils import (
    format_messages_for_tracing,
    get_completion_choices,
    get_finish_reason,
    get_request_parameters,
    get_stream_text,
    get_usage_from_response,
)


class TestGetFinishReason:
    def test_response_with_stop_reason(self):
        mock_response = Mock()
        mock_response.stop_reason = "end_turn"
        assert get_finish_reason(mock_response) == "end_turn"

    def test_response_without_stop_reason(self):
        mock_response = Mock()
        del mock_response.stop_reason
        assert get_finish_reason(mock_response) is None

    def test_none_response(self):
        assert get_finish_reason(None) is None

    def test_response_with_max_tokens_stop_reason(self):
        mock_response = Mock()
        mock_response.stop_reason = "max_tokens"
        assert get_finish_reason(mock_response) == "max_tokens"


class TestGetCompletionChoices:
    def test_response_with_text_content(self):
        mock_content_block = Mock()
        mock_content_block.text = "Hello, world!"
        mock_content_block.type = "text"

        mock_response = Mock()
        mock_response.content = [mock_content_block]
        mock_response.role = "assistant"

        choices = get_completion_choices(mock_response)
        assert len(choices) == 1
        assert choices[0]["message"]["content"] == "Hello, world!"
        assert choices[0]["message"]["role"] == "assistant"
        assert choices[0]["text"] == "Hello, world!"

    def test_response_with_tool_use(self):
        mock_tool_block = Mock()
        mock_tool_block.type = "tool_use"
        mock_tool_block.id = "tool_123"
        mock_tool_block.name = "get_weather"
        mock_tool_block.input = {"location": "San Francisco"}

        mock_response = Mock()
        mock_response.content = [mock_tool_block]
        mock_response.role = "assistant"

        choices = get_completion_choices(mock_response)
        assert len(choices) == 1
        assert choices[0]["message"]["tool_calls"] is not None
        assert len(choices[0]["message"]["tool_calls"]) == 1
        assert choices[0]["message"]["tool_calls"][0]["id"] == "tool_123"
        assert choices[0]["message"]["tool_calls"][0]["function"]["name"] == "get_weather"

        import json
        arguments = choices[0]["message"]["tool_calls"][0]["function"]["arguments"]
        assert isinstance(arguments, str)
        parsed_args = json.loads(arguments)
        assert parsed_args == {"location": "San Francisco"}

    def test_response_without_content(self):
        mock_response = Mock()
        mock_response.content = []
        assert get_completion_choices(mock_response) == []

    def test_none_response(self):
        assert get_completion_choices(None) == []

    def test_response_with_multiple_content_blocks(self):
        mock_text_block = Mock()
        mock_text_block.text = "Here's the weather:"
        mock_text_block.type = "text"

        mock_tool_block = Mock()
        mock_tool_block.type = "tool_use"
        mock_tool_block.id = "tool_456"
        mock_tool_block.name = "get_weather"
        mock_tool_block.input = {"location": "NYC"}

        mock_response = Mock()
        mock_response.content = [mock_text_block, mock_tool_block]
        mock_response.role = "assistant"

        choices = get_completion_choices(mock_response)
        assert len(choices) == 2


class TestGetUsageFromResponse:
    def test_response_with_usage(self):
        mock_usage = Mock()
        mock_usage.input_tokens = 100
        mock_usage.output_tokens = 50

        mock_response = Mock()
        mock_response.usage = mock_usage

        usage = get_usage_from_response(mock_response)
        assert usage["input_tokens"] == 100
        assert usage["output_tokens"] == 50
        assert usage["prompt_tokens"] == 100
        assert usage["completion_tokens"] == 50
        assert usage["total_tokens"] == 150

    def test_response_with_partial_usage(self):
        mock_usage = Mock()
        mock_usage.input_tokens = 100
        del mock_usage.output_tokens

        mock_response = Mock()
        mock_response.usage = mock_usage

        usage = get_usage_from_response(mock_response)
        assert usage["input_tokens"] == 100
        assert usage["output_tokens"] is None
        assert usage["total_tokens"] is None

    def test_response_without_usage(self):
        mock_response = Mock()
        del mock_response.usage
        assert get_usage_from_response(mock_response) == {}

    def test_none_response(self):
        assert get_usage_from_response(None) == {}


class TestGetRequestParameters:
    def test_basic_parameters(self):
        kwargs = {
            "max_tokens": 1024,
            "temperature": 0.7,
            "top_p": 0.9,
            "stream": True,
        }
        params = get_request_parameters(kwargs)
        assert params["max_tokens"] == 1024
        assert params["temperature"] == 0.7
        assert params["top_p"] == 0.9
        assert params["stream"] is True

    def test_missing_parameters(self):
        kwargs = {"max_tokens": 512}
        params = get_request_parameters(kwargs)
        assert params["max_tokens"] == 512
        assert params["temperature"] is None
        assert params["top_p"] is None
        assert params["stream"] is False

    def test_empty_kwargs(self):
        params = get_request_parameters({})
        assert params["max_tokens"] is None
        assert params["temperature"] is None
        assert params["top_p"] is None
        assert params["stream"] is False


class TestFormatMessagesForTracing:
    def test_simple_messages(self):
        messages = [
            {"role": "user", "content": "Hello"},
            {"role": "assistant", "content": "Hi there!"},
        ]
        formatted = format_messages_for_tracing(messages)
        assert len(formatted) == 2
        assert formatted[0]["role"] == "user"
        assert formatted[0]["content"] == "Hello"
        assert formatted[1]["role"] == "assistant"
        assert formatted[1]["content"] == "Hi there!"

    def test_messages_with_content_blocks(self):
        messages = [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": "What's the weather?"},
                ],
            }
        ]
        formatted = format_messages_for_tracing(messages)
        assert len(formatted) == 1
        assert formatted[0]["content"] == "What's the weather?"

    def test_messages_with_tool_results(self):
        messages = [
            {
                "role": "user",
                "content": [
                    {"type": "tool_result", "content": "Sunny, 72°F"},
                ],
            }
        ]
        formatted = format_messages_for_tracing(messages)
        assert len(formatted) == 1
        assert "[Tool Result:" in formatted[0]["content"]

    def test_empty_messages(self):
        assert format_messages_for_tracing([]) == []

    def test_messages_with_mixed_content(self):
        messages = [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": "First part"},
                    {"type": "text", "text": "Second part"},
                ],
            }
        ]
        formatted = format_messages_for_tracing(messages)
        assert len(formatted) == 1
        assert "First part\nSecond part" in formatted[0]["content"]


class TestGetStreamText:
    def test_chunk_with_delta_text(self):
        mock_delta = Mock()
        mock_delta.text = "Hello"

        mock_chunk = Mock()
        mock_chunk.delta = mock_delta

        assert get_stream_text(mock_chunk) == "Hello"

    def test_chunk_with_content_block_text(self):
        mock_content_block = Mock()
        mock_content_block.text = "World"

        mock_chunk = Mock()
        del mock_chunk.delta
        mock_chunk.content_block = mock_content_block

        assert get_stream_text(mock_chunk) == "World"

    def test_chunk_with_none_text(self):
        mock_delta = Mock()
        mock_delta.text = None

        mock_chunk = Mock()
        mock_chunk.delta = mock_delta

        assert get_stream_text(mock_chunk) == ""

    def test_chunk_without_text(self):
        mock_chunk = Mock()
        del mock_chunk.delta
        del mock_chunk.content_block

        assert get_stream_text(mock_chunk) == ""

    def test_none_chunk(self):
        assert get_stream_text(None) == ""


class TestIntegrationScenarios:
    def test_complete_response_processing(self):
        mock_content_block = Mock()
        mock_content_block.text = "The weather is sunny."
        mock_content_block.type = "text"

        mock_usage = Mock()
        mock_usage.input_tokens = 50
        mock_usage.output_tokens = 25

        mock_response = Mock()
        mock_response.content = [mock_content_block]
        mock_response.role = "assistant"
        mock_response.stop_reason = "end_turn"
        mock_response.usage = mock_usage
        mock_response.model = "claude-3-5-sonnet-20241022"
        mock_response.id = "msg_123"

        choices = get_completion_choices(mock_response)
        usage = get_usage_from_response(mock_response)
        finish_reason = get_finish_reason(mock_response)

        assert len(choices) == 1
        assert choices[0]["message"]["content"] == "The weather is sunny."
        assert usage["total_tokens"] == 75
        assert finish_reason == "end_turn"

    def test_streaming_chunk_processing(self):
        mock_delta = Mock()
        mock_delta.text = "Streaming"

        mock_chunk = Mock()
        mock_chunk.delta = mock_delta
        mock_chunk.type = "content_block_delta"

        text = get_stream_text(mock_chunk)
        assert text == "Streaming"



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import SpanProcessor


class Exporter(ABC):
    @abstractmethod
    def __init__(self, config: Any):
        raise NotImplementedError("`__init__` is not implemented")

    @abstractmethod
    def create_span_processor(self, resource: Resource) -> SpanProcessor:
        raise NotImplementedError("`create_span_processor` is not implemented")



from verse_sdk.spec.utils import is_valid_type


class TestIsValidType:
    def test_string_type_valid(self):
        assert is_valid_type("hello", "string") is True
        assert is_valid_type("", "string") is True

    def test_string_type_invalid(self):
        assert is_valid_type(123, "string") is False
        assert is_valid_type(12.5, "string") is False
        assert is_valid_type(True, "string") is False
        assert is_valid_type(["hello"], "string") is False

    def test_int_type_valid(self):
        assert is_valid_type(123, "int") is True
        assert is_valid_type(0, "int") is True
        assert is_valid_type(-456, "int") is True

    def test_int_type_invalid(self):
        assert is_valid_type("123", "int") is False
        assert is_valid_type(12.5, "int") is False
        assert is_valid_type(True, "int") is False  # bool is subclass of int in Python
        assert is_valid_type(False, "int") is False

    def test_double_type_valid(self):
        assert is_valid_type(12.5, "double") is True
        assert is_valid_type(123, "double") is True  # int is accepted for double
        assert is_valid_type(0.0, "double") is True
        assert is_valid_type(-45.67, "double") is True

    def test_double_type_invalid(self):
        assert is_valid_type("12.5", "double") is False
        assert is_valid_type(True, "double") is False
        assert is_valid_type(False, "double") is False

    def test_boolean_type_valid(self):
        assert is_valid_type(True, "boolean") is True
        assert is_valid_type(False, "boolean") is True

    def test_boolean_type_invalid(self):
        assert is_valid_type(1, "boolean") is False
        assert is_valid_type(0, "boolean") is False
        assert is_valid_type("true", "boolean") is False

    def test_string_array_type_valid(self):
        assert is_valid_type(["hello", "world"], "string[]") is True
        assert is_valid_type([], "string[]") is True
        assert is_valid_type(["single"], "string[]") is True

    def test_string_array_type_invalid(self):
        assert is_valid_type("hello", "string[]") is False
        assert is_valid_type([1, 2, 3], "string[]") is False
        assert is_valid_type(["hello", 123], "string[]") is False
        assert is_valid_type([True, False], "string[]") is False

    def test_any_type_accepts_all(self):
        assert is_valid_type("hello", "any") is True
        assert is_valid_type(123, "any") is True
        assert is_valid_type(12.5, "any") is True
        assert is_valid_type(True, "any") is True
        assert is_valid_type(["hello"], "any") is True
        assert is_valid_type({"key": "value"}, "any") is True
        assert is_valid_type(None, "any") is True

    def test_unknown_type_returns_false(self):
        assert is_valid_type("hello", "unknown_type") is False
        assert is_valid_type(123, "custom") is False



from unittest.mock import Mock

from pydantic_ai import RunUsage

from verse_sdk.integrations.pydantic_ai.utils import (
    calculate_total_tokens,
    get_finish_reason,
    get_generation_messages,
    get_operation_type,
    get_usage,
)


class TestCalculateTotalTokens:
    def test_calculate_total_tokens_with_valid_usage(self):
        assert (
            calculate_total_tokens(
                RunUsage(
                    input_tokens=100,
                    output_tokens=50,
                    cache_read_tokens=0,
                    requests=1,
                    tool_calls=0,
                    details={},
                )
            )
            == 150
        )

    def test_calculate_total_tokens_with_zero_tokens(self):
        assert (
            calculate_total_tokens(
                RunUsage(
                    input_tokens=0,
                    output_tokens=0,
                    cache_read_tokens=0,
                    requests=1,
                    tool_calls=0,
                    details={},
                )
            )
            is None
        )

    def test_calculate_total_tokens_with_none_input_tokens(self):
        assert (
            calculate_total_tokens(
                RunUsage(
                    input_tokens=None,
                    output_tokens=50,
                    cache_read_tokens=0,
                    requests=1,
                    tool_calls=0,
                    details={},
                )
            )
            is None
        )

    def test_calculate_total_tokens_with_none_output_tokens(self):
        assert (
            calculate_total_tokens(
                RunUsage(
                    input_tokens=100,
                    output_tokens=None,
                    cache_read_tokens=0,
                    requests=1,
                    tool_calls=0,
                    details={},
                )
            )
            is None
        )


class TestGetFinishReason:
    def test_get_finish_reason_with_messages(self):
        mock_message = Mock()
        mock_message.finish_reason = "stop"

        mock_ctx = Mock()
        mock_ctx.messages = [mock_message]
        assert get_finish_reason(mock_ctx) == "stop"

    def test_get_finish_reason_with_multiple_messages(self):
        mock_message1 = Mock()
        mock_message1.finish_reason = "length"
        mock_message2 = Mock()
        mock_message2.finish_reason = "tool_call"

        mock_ctx = Mock()
        mock_ctx.messages = [mock_message1, mock_message2]
        assert get_finish_reason(mock_ctx) == "tool_call"

    def test_get_finish_reason_with_no_messages(self):
        mock_ctx = Mock()
        mock_ctx.messages = []
        assert get_finish_reason(mock_ctx) == ""

    def test_get_finish_reason_with_no_messages_attribute(self):
        mock_ctx = Mock()
        del mock_ctx.messages
        assert get_finish_reason(mock_ctx) == ""

    def test_get_finish_reason_with_none_messages(self):
        mock_ctx = Mock()
        mock_ctx.messages = None
        assert get_finish_reason(mock_ctx) == ""

    def test_get_finish_reason_with_empty_finish_reason(self):
        mock_message = Mock()
        mock_message.finish_reason = ""

        mock_ctx = Mock()
        mock_ctx.messages = [mock_message]
        assert get_finish_reason(mock_ctx) == ""


class TestGetGenerationMessages:
    def test_get_generation_messages_with_no_messages(self):
        mock_ctx = Mock()
        mock_ctx.messages = []
        prompts, completions = get_generation_messages(mock_ctx)
        assert prompts == []
        assert completions == []

    def test_get_generation_messages_with_no_messages_attribute(self):
        mock_ctx = Mock()
        del mock_ctx.messages
        prompts, completions = get_generation_messages(mock_ctx)
        assert prompts == []
        assert completions == []

    def test_get_generation_messages_with_system_prompt(self):
        mock_part = Mock()
        mock_part.__class__.__name__ = "SystemPromptPart"
        mock_part.content = "You are a helpful assistant"

        mock_message = Mock()
        mock_message.parts = [mock_part]

        mock_ctx = Mock()
        mock_ctx.messages = [mock_message]
        prompts, completions = get_generation_messages(mock_ctx)

        assert completions == []
        assert len(prompts) == 1
        assert prompts[0]["role"] == "system"
        assert prompts[0]["content"] == "You are a helpful assistant"
        assert prompts[0]["function_call"] is None
        assert prompts[0]["name"] is None
        assert prompts[0]["tool_calls"] is None

    def test_get_generation_messages_with_user_prompt(self):
        mock_part = Mock()
        mock_part.__class__.__name__ = "UserPromptPart"
        mock_part.content = "Hello, how are you?"

        mock_message = Mock()
        mock_message.parts = [mock_part]

        mock_ctx = Mock()
        mock_ctx.messages = [mock_message]

        prompts, completions = get_generation_messages(mock_ctx)

        assert completions == []
        assert len(prompts) == 1
        assert prompts[0]["role"] == "user"
        assert prompts[0]["content"] == "Hello, how are you?"
        assert prompts[0]["function_call"] is None
        assert prompts[0]["name"] is None
        assert prompts[0]["tool_calls"] is None

    def test_get_generation_messages_with_tool_call(self):
        mock_part = Mock()
        mock_part.__class__.__name__ = "ToolCallPart"
        mock_part.tool_name = "get_weather"
        mock_part.args = '{"location": "New York"}'
        mock_part.tool_call_id = "call_123"

        mock_message = Mock()
        mock_message.parts = [mock_part]

        mock_ctx = Mock()
        mock_ctx.messages = [mock_message]
        prompts, completions = get_generation_messages(mock_ctx)

        assert prompts == []
        assert len(completions) == 1
        completion_message = completions[0]["message"]
        assert completion_message["role"] == "assistant"
        assert completion_message["content"] == ""
        assert completion_message["function_call"] is None
        assert completion_message["name"] is None
        assert len(completion_message["tool_calls"]) == 1
        assert completion_message["tool_calls"][0]["id"] == "call_123"
        assert completion_message["tool_calls"][0]["type"] == "function"
        assert completion_message["tool_calls"][0]["function"]["name"] == "get_weather"
        assert (
            completion_message["tool_calls"][0]["function"]["arguments"]
            == '{"location": "New York"}'
        )

    def test_get_generation_messages_with_text_part(self):
        mock_part = Mock()
        mock_part.__class__.__name__ = "TextPart"
        mock_part.content = "I can help you with that."

        mock_message = Mock()
        mock_message.parts = [mock_part]

        mock_ctx = Mock()
        mock_ctx.messages = [mock_message]
        prompts, completions = get_generation_messages(mock_ctx)

        assert prompts == []
        assert len(completions) == 1
        completion_message = completions[0]["message"]
        assert completion_message["role"] == "assistant"
        assert completion_message["content"] == "I can help you with that."
        assert completion_message["function_call"] is None
        assert completion_message["name"] is None
        assert completion_message["tool_calls"] is None

    def test_get_generation_messages_with_tool_return_part(self):
        mock_part = Mock()
        mock_part.__class__.__name__ = "ToolReturnPart"
        mock_part.content = "Tool result"
        mock_part.tool_call_id = "call_123"

        mock_message = Mock()
        mock_message.parts = [mock_part]

        mock_ctx = Mock()
        mock_ctx.messages = [mock_message]

        prompts, completions = get_generation_messages(mock_ctx)

        assert completions == []
        assert prompts == [
            {
                "role": "tool",
                "content": '"Tool result"',
                "function_call": None,
                "name": None,
                "tool_calls": None,
                "id": "call_123",
            }
        ]

    def test_get_generation_messages_with_multiple_parts(self):
        mock_text_part = Mock()
        mock_text_part.__class__.__name__ = "TextPart"
        mock_text_part.content = "I'll help you with that."

        mock_tool_part = Mock()
        mock_tool_part.__class__.__name__ = "ToolCallPart"
        mock_tool_part.tool_name = "search"
        mock_tool_part.args = '{"query": "test"}'
        mock_tool_part.tool_call_id = "call_456"

        mock_message = Mock()
        mock_message.parts = [mock_text_part, mock_tool_part]

        mock_ctx = Mock()
        mock_ctx.messages = [mock_message]

        prompts, completions = get_generation_messages(mock_ctx)

        assert prompts == []
        assert len(completions) == 1
        completion_message = completions[0]["message"]
        assert completion_message["role"] == "assistant"
        assert completion_message["content"] == "I'll help you with that."
        assert len(completion_message["tool_calls"]) == 1

    def test_get_generation_messages_with_multiple_messages(self):
        mock_system_part = Mock()
        mock_system_part.__class__.__name__ = "SystemPromptPart"
        mock_system_part.content = "You are helpful"

        mock_user_part = Mock()
        mock_user_part.__class__.__name__ = "UserPromptPart"
        mock_user_part.content = "Hello"

        mock_system_message = Mock()
        mock_system_message.parts = [mock_system_part]

        mock_user_message = Mock()
        mock_user_message.parts = [mock_user_part]

        mock_ctx = Mock()
        mock_ctx.messages = [mock_system_message, mock_user_message]

        prompts, completions = get_generation_messages(mock_ctx)

        assert completions == []
        assert len(prompts) == 2
        assert prompts[0]["role"] == "system"
        assert prompts[0]["content"] == "You are helpful"
        assert prompts[1]["role"] == "user"
        assert prompts[1]["content"] == "Hello"

    def test_get_generation_messages_with_message_without_parts(self):
        mock_message = Mock()
        del mock_message.parts

        mock_ctx = Mock()
        mock_ctx.messages = [mock_message]

        prompts, completions = get_generation_messages(mock_ctx)
        assert prompts == []
        assert completions == []


class TestGetOperationType:
    def test_get_operation_type_with_messages(self):
        mock_ctx = Mock()
        mock_ctx.messages = [Mock()]
        assert get_operation_type(mock_ctx) == "chat"

    def test_get_operation_type_with_no_messages(self):
        mock_ctx = Mock()
        mock_ctx.messages = []
        assert get_operation_type(mock_ctx) == "completion"

    def test_get_operation_type_with_no_messages_attribute(self):
        mock_ctx = Mock()
        del mock_ctx.messages
        assert get_operation_type(mock_ctx) == "completion"

    def test_get_operation_type_with_none_messages(self):
        mock_ctx = Mock()
        mock_ctx.messages = None
        assert get_operation_type(mock_ctx) == "completion"


class TestGetUsage:
    def test_get_usage_with_valid_data(self):
        usage = RunUsage(
            input_tokens=100,
            output_tokens=50,
            cache_read_tokens=25,
            requests=1,
            tool_calls=2,
            details={},
        )

        result = get_usage(usage)
        assert result["completion_tokens"] == 50
        assert result["input_tokens"] == 100
        assert result["output_tokens"] == 50
        assert result["prompt_tokens"] == 100
        assert result["tool_calls"] == 2
        assert result["total_tokens"] == 150

    def test_get_usage_with_none_tokens(self):
        usage = RunUsage(
            input_tokens=None,
            output_tokens=None,
            cache_read_tokens=0,
            requests=1,
            tool_calls=0,
            details={},
        )

        result = get_usage(usage)
        assert result["completion_tokens"] is None
        assert result["input_tokens"] is None
        assert result["output_tokens"] is None
        assert result["prompt_tokens"] is None
        assert result["tool_calls"] == 0
        assert result["total_tokens"] is None

    def test_get_usage_with_zero_tokens(self):
        usage = RunUsage(
            input_tokens=0,
            output_tokens=0,
            cache_read_tokens=0,
            requests=1,
            tool_calls=0,
            details={},
        )

        result = get_usage(usage)
        assert result["completion_tokens"] == 0
        assert result["input_tokens"] == 0
        assert result["output_tokens"] == 0
        assert result["prompt_tokens"] == 0
        assert result["tool_calls"] == 0
        assert result["total_tokens"] is None



from unittest.mock import Mock

from verse_sdk.integrations.google.utils import (
    format_messages_for_tracing,
    get_completion_choices,
    get_finish_reason,
    get_request_parameters,
    get_stream_text,
    get_usage_from_response,
)


class TestGetFinishReason:
    def test_response_with_stop_finish_reason(self):
        mock_candidate = Mock()
        mock_candidate.finish_reason = "STOP"

        mock_response = Mock()
        mock_response.candidates = [mock_candidate]

        assert get_finish_reason(mock_response) == "stop"

    def test_response_with_max_tokens_finish_reason(self):
        mock_candidate = Mock()
        mock_candidate.finish_reason = "MAX_TOKENS"

        mock_response = Mock()
        mock_response.candidates = [mock_candidate]

        assert get_finish_reason(mock_response) == "length"

    def test_response_with_safety_finish_reason(self):
        mock_candidate = Mock()
        mock_candidate.finish_reason = "SAFETY"

        mock_response = Mock()
        mock_response.candidates = [mock_candidate]

        assert get_finish_reason(mock_response) == "content_filter"

    def test_response_with_other_finish_reason(self):
        mock_candidate = Mock()
        mock_candidate.finish_reason = "OTHER"

        mock_response = Mock()
        mock_response.candidates = [mock_candidate]

        assert get_finish_reason(mock_response) == "other"

    def test_response_without_candidates(self):
        mock_response = Mock()
        mock_response.candidates = []
        assert get_finish_reason(mock_response) is None

    def test_none_response(self):
        assert get_finish_reason(None) is None


class TestGetCompletionChoices:
    def test_response_with_text_content(self):
        mock_part = Mock()
        mock_part.text = "Hello, world!"

        mock_content = Mock()
        mock_content.parts = [mock_part]
        mock_content.role = "model"

        mock_candidate = Mock()
        mock_candidate.content = mock_content

        mock_response = Mock()
        mock_response.candidates = [mock_candidate]

        choices = get_completion_choices(mock_response)
        assert len(choices) == 1
        assert choices[0]["message"]["content"] == "Hello, world!"
        assert choices[0]["message"]["role"] == "model"
        assert choices[0]["text"] == "Hello, world!"

    def test_response_with_function_call(self):
        mock_function_call = Mock()
        mock_function_call.name = "get_weather"
        mock_function_call.args = {"location": "San Francisco"}

        mock_part = Mock()
        del mock_part.text
        mock_part.function_call = mock_function_call

        mock_content = Mock()
        mock_content.parts = [mock_part]
        mock_content.role = "model"

        mock_candidate = Mock()
        mock_candidate.content = mock_content

        mock_response = Mock()
        mock_response.candidates = [mock_candidate]

        choices = get_completion_choices(mock_response)
        assert len(choices) == 1
        assert choices[0]["message"]["tool_calls"] is not None
        assert len(choices[0]["message"]["tool_calls"]) == 1
        assert choices[0]["message"]["tool_calls"][0]["function"]["name"] == "get_weather"

    def test_response_without_candidates(self):
        mock_response = Mock()
        mock_response.candidates = []
        assert get_completion_choices(mock_response) == []

    def test_none_response(self):
        assert get_completion_choices(None) == []

    def test_response_with_multiple_text_parts(self):
        mock_part1 = Mock()
        mock_part1.text = "Hello, "

        mock_part2 = Mock()
        mock_part2.text = "world!"

        mock_content = Mock()
        mock_content.parts = [mock_part1, mock_part2]
        mock_content.role = "model"

        mock_candidate = Mock()
        mock_candidate.content = mock_content

        mock_response = Mock()
        mock_response.candidates = [mock_candidate]

        choices = get_completion_choices(mock_response)
        assert len(choices) == 1
        assert choices[0]["message"]["content"] == "Hello, world!"


class TestGetUsageFromResponse:
    def test_response_with_usage(self):
        mock_usage = Mock()
        mock_usage.prompt_token_count = 100
        mock_usage.response_token_count = 50
        mock_usage.total_token_count = 150

        mock_response = Mock()
        mock_response.usage_metadata = mock_usage

        usage = get_usage_from_response(mock_response)
        assert usage["input_tokens"] == 100
        assert usage["output_tokens"] == 50
        assert usage["prompt_tokens"] == 100
        assert usage["completion_tokens"] == 50
        assert usage["total_tokens"] == 150

    def test_response_with_partial_usage(self):
        mock_usage = Mock()
        mock_usage.prompt_token_count = 100
        del mock_usage.response_token_count
        del mock_usage.total_token_count

        mock_response = Mock()
        mock_response.usage_metadata = mock_usage

        usage = get_usage_from_response(mock_response)
        assert usage["input_tokens"] == 100
        assert usage["output_tokens"] is None
        assert usage["total_tokens"] is None

    def test_response_without_usage(self):
        mock_response = Mock()
        del mock_response.usage_metadata
        assert get_usage_from_response(mock_response) == {}

    def test_none_response(self):
        assert get_usage_from_response(None) == {}


class TestGetRequestParameters:
    def test_basic_parameters(self):
        mock_config = Mock()
        mock_config.max_output_tokens = 1024
        mock_config.temperature = 0.7
        mock_config.top_p = 0.9
        mock_config.top_k = 40

        kwargs = {
            "config": mock_config,
            "stream": True,
        }

        params = get_request_parameters(kwargs)
        assert params["max_tokens"] == 1024
        assert params["temperature"] == 0.7
        assert params["top_p"] == 0.9
        assert params["stream"] is False

    def test_missing_parameters(self):
        mock_config = Mock()
        mock_config.max_output_tokens = 512
        del mock_config.temperature
        del mock_config.top_p
        del mock_config.top_k

        kwargs = {"config": mock_config}

        params = get_request_parameters(kwargs)
        assert params["max_tokens"] == 512
        assert params["temperature"] is None
        assert params["top_p"] is None
        assert params["stream"] is False

    def test_empty_kwargs(self):
        params = get_request_parameters({})
        assert params["max_tokens"] is None
        assert params["temperature"] is None
        assert params["top_p"] is None
        assert params["stream"] is False


class TestFormatMessagesForTracing:
    def test_simple_string_content(self):
        contents = "Hello, how are you?"
        formatted = format_messages_for_tracing(contents)
        assert len(formatted) == 1
        assert formatted[0]["role"] == "user"
        assert formatted[0]["content"] == "Hello, how are you?"

    def test_list_of_strings(self):
        contents = ["Hello", "How are you?"]
        formatted = format_messages_for_tracing(contents)
        assert len(formatted) == 2
        assert formatted[0]["content"] == "Hello"
        assert formatted[1]["content"] == "How are you?"

    def test_dict_with_role_and_parts(self):
        contents = [
            {
                "role": "user",
                "parts": [{"text": "What's the weather?"}],
            }
        ]
        formatted = format_messages_for_tracing(contents)
        assert len(formatted) == 1
        assert formatted[0]["role"] == "user"
        assert formatted[0]["content"] == "What's the weather?"

    def test_dict_with_function_call(self):
        contents = [
            {
                "role": "model",
                "parts": [{"function_call": {"name": "get_weather"}}],
            }
        ]
        formatted = format_messages_for_tracing(contents)
        assert len(formatted) == 1
        assert "[Function Call: get_weather]" in formatted[0]["content"]

    def test_object_with_role_and_parts(self):
        mock_part = Mock()
        mock_part.text = "Hello from object"

        mock_content = Mock()
        mock_content.role = "user"
        mock_content.parts = [mock_part]

        formatted = format_messages_for_tracing(mock_content)
        assert len(formatted) == 1
        assert formatted[0]["role"] == "user"
        assert formatted[0]["content"] == "Hello from object"

    def test_empty_contents(self):
        assert format_messages_for_tracing([]) == []

    def test_mixed_text_parts(self):
        contents = [
            {
                "role": "user",
                "parts": [
                    {"text": "First part"},
                    {"text": "Second part"},
                ],
            }
        ]
        formatted = format_messages_for_tracing(contents)
        assert len(formatted) == 1
        assert "First part\nSecond part" in formatted[0]["content"]


class TestGetStreamText:
    def test_chunk_with_text_attribute(self):
        mock_chunk = Mock()
        mock_chunk.text = "Hello"
        assert get_stream_text(mock_chunk) == "Hello"

    def test_chunk_with_candidates_and_parts(self):
        mock_part = Mock()
        mock_part.text = "World"

        mock_content = Mock()
        mock_content.parts = [mock_part]

        mock_candidate = Mock()
        mock_candidate.content = mock_content

        mock_chunk = Mock()
        del mock_chunk.text
        mock_chunk.candidates = [mock_candidate]

        assert get_stream_text(mock_chunk) == "World"

    def test_chunk_with_multiple_parts(self):
        mock_part1 = Mock()
        mock_part1.text = "Hello "

        mock_part2 = Mock()
        mock_part2.text = "World"

        mock_content = Mock()
        mock_content.parts = [mock_part1, mock_part2]

        mock_candidate = Mock()
        mock_candidate.content = mock_content

        mock_chunk = Mock()
        del mock_chunk.text
        mock_chunk.candidates = [mock_candidate]

        assert get_stream_text(mock_chunk) == "Hello World"

    def test_chunk_without_text(self):
        mock_chunk = Mock()
        del mock_chunk.text
        mock_chunk.candidates = []
        assert get_stream_text(mock_chunk) == ""

    def test_none_chunk(self):
        assert get_stream_text(None) == ""


class TestIntegrationScenarios:
    def test_complete_response_processing(self):
        mock_part = Mock()
        mock_part.text = "The weather is sunny."

        mock_content = Mock()
        mock_content.parts = [mock_part]
        mock_content.role = "model"

        mock_candidate = Mock()
        mock_candidate.content = mock_content
        mock_candidate.finish_reason = "STOP"

        mock_usage = Mock()
        mock_usage.prompt_token_count = 50
        mock_usage.response_token_count = 25
        mock_usage.total_token_count = 75

        mock_response = Mock()
        mock_response.candidates = [mock_candidate]
        mock_response.usage_metadata = mock_usage

        choices = get_completion_choices(mock_response)
        usage = get_usage_from_response(mock_response)
        finish_reason = get_finish_reason(mock_response)

        assert len(choices) == 1
        assert choices[0]["message"]["content"] == "The weather is sunny."
        assert usage["total_tokens"] == 75
        assert finish_reason == "stop"

    def test_streaming_chunk_processing(self):
        mock_part = Mock()
        mock_part.text = "Streaming"

        mock_content = Mock()
        mock_content.parts = [mock_part]

        mock_candidate = Mock()
        mock_candidate.content = mock_content

        mock_chunk = Mock()
        del mock_chunk.text
        mock_chunk.candidates = [mock_candidate]

        text = get_stream_text(mock_chunk)
        assert text == "Streaming"



from __future__ import annotations

import logging
from collections.abc import Iterable
from typing import Any

from ...contexts import CompletionChoice, GenerationMessage, GenerationUsage
from ...utils import get

# LangChain-specific vendor list for detecting LLM providers from serialized metadata.
# When LangChain serializes components, it uses an "id" field with a list structure:
# e.g., {"id": ["langchain", "openai", "chat"]} becomes "langchain/openai/chat"
# We then check if any of these vendor names appear in that serialized string.
# This approach is specific to LangChain's metadata format and differs from how
# other integrations (Pydantic AI, LiteLLM) handle vendor identification.
KNOWN_VENDORS = (
    "openai",
    "anthropic",
    "google",
    "groq",
    "mistral",
    "ollama",
    "cohere",
)


def _serialized_to_str(serialized: Any) -> str:
    if isinstance(serialized, dict) and "id" in serialized:
        identifier = serialized["id"]
        if isinstance(identifier, (list, tuple)):
            return "/".join(str(part) for part in identifier).lower()
        return str(identifier).lower()
    return str(serialized).lower()


def get_vendor_and_model(serialized: Any, **kwargs: Any) -> tuple[str, str]:
    """
    Best-effort vendor/model extraction based on LangChain metadata.

    LangChain serializes components with an "id" field containing a list:
    - {"id": ["langchain", "openai", "chat"]} -> detects "openai"
    - {"id": ["langchain", "anthropic", "claude"]} -> detects "anthropic"
    - {"id": ["langchain", "custom", "llm"]} -> returns "unknown"

    The detection works by converting the list to a string (e.g., "langchain/openai/chat")
    and checking if any KNOWN_VENDORS appear as substrings.
    """
    serialized_str = _serialized_to_str(serialized)
    vendor = next((v for v in KNOWN_VENDORS if v in serialized_str), "unknown")

    if vendor == "unknown" and serialized_str:
        logging.debug(
            f"Could not identify LangChain vendor from serialized metadata: {serialized_str}. "
        )

    invocation_params = kwargs.get("invocation_params", {}) or {}
    model = (
        get(invocation_params, "model")
        or get(invocation_params, "model_name")
        or get(kwargs, "model")
        or get(kwargs, "model_name")
        or ""
    )

    return vendor, model or "unknown"


def _role_from_message_obj(message: Any) -> str:
    explicit_role = getattr(message, "role", None) or getattr(message, "type", None)
    if explicit_role:
        role = str(explicit_role).lower()
        if role in {"assistant", "ai"}:
            return "assistant"
        if role in {"human", "user"}:
            return "user"
        if role in {"system"}:
            return "system"
        if role in {"tool", "function"}:
            return "tool"
        return role

    name = message.__class__.__name__.lower()
    if "system" in name:
        return "system"
    if "human" in name or "user" in name:
        return "user"
    if "ai" in name or "assistant" in name:
        return "assistant"
    if "tool" in name or "function" in name:
        return "tool"
    return "user"


def _text_from_content(content: Any) -> str:
    if isinstance(content, list):
        parts: list[str] = []
        for part in content:
            text = get(part, "text") or get(part, "content")
            parts.append(str(text) if text is not None else str(part))
        return "\n".join(parts)
    return str(content) if content is not None else ""


def get_generation_messages_from_chat(
    messages_batches: Iterable[Iterable[Any]],
) -> list[GenerationMessage]:
    prompts: list[GenerationMessage] = []
    for batch in messages_batches or []:
        for message in batch or []:
            prompts.append(
                GenerationMessage(
                    role=_role_from_message_obj(message),
                    content=_text_from_content(get(message, "content")),
                    function_call=None,
                    name=None,
                    tool_calls=None,
                )
            )

    return prompts


def get_generation_messages_from_prompts(
    prompts: Iterable[str],
) -> list[GenerationMessage]:
    prompt_messages: list[GenerationMessage] = []
    for prompt in prompts or []:
        prompt_messages.append(
            GenerationMessage(
                role="user",
                content=str(prompt),
                function_call=None,
                name=None,
                tool_calls=None,
            )
        )
    return prompt_messages


def _tool_calls_from_message(message: Any) -> list[dict] | None:
    tool_calls = getattr(message, "tool_calls", None)
    if not tool_calls:
        additional = getattr(message, "additional_kwargs", None)
        if isinstance(additional, dict):
            tool_calls = additional.get("tool_calls")

    if isinstance(tool_calls, list) and tool_calls:
        return tool_calls
    return None


def _content_from_generation(generation: Any) -> str:
    """Extract text content from a generation object (handles both text and chat formats)."""
    if hasattr(generation, "text") and generation.text is not None:
        return str(generation.text)

    message = getattr(generation, "message", None)
    if message is not None:
        return _text_from_content(get(message, "content"))

    return ""


def _finish_reason_from_generation(generation: Any) -> str | None:
    generation_info = getattr(generation, "generation_info", None)
    if isinstance(generation_info, dict) and generation_info.get("finish_reason"):
        return str(generation_info["finish_reason"])

    message = getattr(generation, "message", None)
    response_metadata = getattr(message, "response_metadata", None)
    if isinstance(response_metadata, dict) and response_metadata.get("finish_reason"):
        return str(response_metadata["finish_reason"])

    return None


def get_completions_from_result(
    result: Any,
) -> tuple[list[CompletionChoice], str | None, str | None]:
    completions: list[CompletionChoice] = []
    finish_reason: str | None = None
    model_used: str | None = None

    generations = getattr(result, "generations", None) or []
    flat_generations: list[Any] = []
    for generation in generations:
        if isinstance(generation, (list, tuple)):
            flat_generations.extend(generation)
        else:
            flat_generations.append(generation)

    for generation in flat_generations:
        message = getattr(generation, "message", None)

        if message is not None:
            completions.append(
                CompletionChoice(
                    message={
                        "role": "assistant",
                        "content": _text_from_content(get(message, "content")),
                        "function_call": None,
                        "name": None,
                        "tool_calls": _tool_calls_from_message(message),
                    },
                    delta=None,
                    text=None,
                )
            )
        else:
            completions.append(
                CompletionChoice(
                    message=None,
                    delta=None,
                    text=_content_from_generation(generation),
                )
            )

        finish_reason = _finish_reason_from_generation(generation) or finish_reason

    llm_output = getattr(result, "llm_output", None)
    if isinstance(llm_output, dict):
        model_used = (
            llm_output.get("model_name")
            or llm_output.get("model")
            or llm_output.get("model_id")
        )

    return completions, finish_reason, model_used


def get_usage_from_result(result: Any) -> GenerationUsage:
    llm_output = getattr(result, "llm_output", None)
    token_usage = None
    if isinstance(llm_output, dict):
        token_usage = llm_output.get("token_usage") or llm_output.get("usage")
    # Fallback: some LC wrappers surface usage here
    token_usage = token_usage or getattr(result, "usage_metadata", None)

    if not isinstance(token_usage, dict):
        return GenerationUsage(
            completion_tokens=None,
            input_tokens=None,
            max_tokens=None,
            output_tokens=None,
            prompt_tokens=None,
            stream=None,
            temperature=None,
            tool_calls=None,
            total_tokens=None,
            top_p=None,
        )

    prompt_tokens = token_usage.get("prompt_tokens")
    completion_tokens = token_usage.get("completion_tokens")
    total_tokens = token_usage.get("total_tokens")

    return GenerationUsage(
        completion_tokens=completion_tokens,
        input_tokens=prompt_tokens,
        max_tokens=None,
        output_tokens=completion_tokens,
        prompt_tokens=prompt_tokens,
        stream=None,
        temperature=None,
        tool_calls=None,
        total_tokens=total_tokens
        if total_tokens is not None
        else (
            (prompt_tokens or 0) + (completion_tokens or 0)
            if prompt_tokens is not None or completion_tokens is not None
            else None
        ),
        top_p=None,
    )



from __future__ import annotations

import json
import logging

from opentelemetry.trace import Span, Status, StatusCode

from .types import ContextMetadata, ContextType, ObservationType, OperationType, Score


class BaseContext:
    def __init__(self, span: Span):
        """Initialize the base observation context."""
        if not span or not span.get_span_context().is_valid:
            logging.error("Context received invalid span", span=span)
            raise ValueError("Span invalid")

        self._span = span
        self._span.set_attribute("context.type", self.type)
        self._span.set_attribute("observation.type", self.observation_type)

    @property
    def span(self) -> Span:
        """Get the underlying OpenTelemetry span."""
        return self._span

    @property
    def observation_type(self) -> ObservationType:
        """Get the type of observation being created."""
        raise NotImplementedError("Observation type must be implemented")

    @property
    def type(self) -> ContextType:
        """Identify the type of context being created."""
        raise NotImplementedError("Context type must be implemented")

    def metadata(self, data: ContextMetadata) -> BaseContext:
        """Add metadata to the span."""
        self.set_attributes(prefix="metadata", **data)
        return self

    def error(self, exception: Exception) -> BaseContext:
        """Record an error on the span."""
        try:
            self._span.set_attribute("error", True)
            self._span.set_attribute("error.type", type(exception).__name__)
            self._span.set_attribute("error.message", str(exception))
            self._span.record_exception(exception)
            self._span.set_status(Status(StatusCode.ERROR, str(exception)))
            return self
        except Exception as e:
            logging.warning("Error setting exception on context", exc_info=e)
            return self

    def get_trace_metadata(self) -> dict:
        """Get the trace metadata for the span (used to connect that traces that are out of bounds)"""
        span_context = self._span.context
        parent_span_id_hex = format(span_context.span_id, "016x")
        trace_id_hex = format(span_context.trace_id, "032x")

        return {
            "parent_observation_id": parent_span_id_hex,
            "trace_id": trace_id_hex,
        }

    def operation(self, operation: OperationType) -> BaseContext:
        """Set the AI operation of the context."""
        self._span.set_attribute("ai.operation", operation)
        self._span.set_attribute("langfuse.observation.type", operation)
        return self

    def score(
        self,
        score: Score,
    ) -> BaseContext:
        """Set score details and register it as an event."""
        try:
            score_type = f"{self.type}_score"
            score_type_key = (
                f"{score_type}.{score['name']}"
                if self.type != "observation"
                else score["name"]
            )

            self.set_attributes(
                prefix=score_type_key,
                comment=score["comment"],
                value=str(score["value"]),
            )

            self._span.add_event(
                score_type,
                {
                    "name": score["name"],
                    "value": str(score["value"]),
                    "comment": score["comment"] or "",
                },
            )

            return self
        except Exception as e:
            logging.warning("Error setting score on context", exc_info=e)
            return self

    def set_attributes(self, prefix: str | None = None, **attributes) -> BaseContext:
        """Helper to set attributes with optional prefix."""
        try:
            for key, value in attributes.items():
                if value is not None:
                    attr_key = f"{prefix}.{key}" if prefix else key

                    # Convert complex types to JSON strings
                    if isinstance(value, (dict, list)):
                        value = json.dumps(value)
                    elif not isinstance(value, (bool, str, bytes, int, float)) and not (
                        isinstance(value, (list, tuple))
                        and all(
                            isinstance(v, (bool, str, bytes, int, float)) for v in value
                        )
                    ):
                        value = str(value)
                    self._span.set_attribute(attr_key, value)

            return self
        except Exception as e:
            logging.warning("Error setting attributes on context", exc_info=e)
            return self

    def status(self, status: Status) -> BaseContext:
        """Set the status for the span."""
        self._span.set_status(status)
        return self
